/*   
 *   TMS9900 C Compiler
 *
 *    0 am_immed  -  immediate mode (e.g. sra r5,8; 8 is the immediate)
 *    1 am_direct -  direct access  (e.g. mov @a,@b, a and b are direct)
 *    2 am_reg    -  register access (e.g. mov r5,r5, r5 is direct )
 *    3 am_ind    -  indirect access (e.g. mov *r5,*r5,  *r5 is indirect)
 *    4 am_ainc   -  autoincrement   (e.g. mov *r5+,*r5+, *r5+ is auto)
 *    5 am_indm   -  indirect memory (e.g. mov @a(r5),@b(r6) )
 *
 *  Copyright 1990 by LGMA Products
 *
 *  This program is free for you to use for non-commercial purposes.
 *  It is, however, copyrighted so that commercial use is strictly
 *  forbidden.
 *
 *  Please contact:
 *
 *     LGMA PRODUCTS
 *     5618 AppleButter Hill Road
 *     Coopersburg, PA  18036
 *
 *  For suggestions, improvements, and the like.
 *
 */

enum e_am {
	am_immed, am_direct, am_reg, am_ind, am_ainc, am_indm };


enum e_op {
	op_a, op_ab, op_abs, op_ai, op_andi, op_b, op_bl, op_blwp,
        op_c, op_cb, op_ci, op_clr, op_coc, op_czc,
        op_dec, op_dect, op_div, op_divs, op_inc, op_inct, op_inv,
        op_jeq, op_jgt, op_jh, op_jhe, op_jl, op_jle, op_jlt, op_jmp,
	op_jnc, op_jne, op_jno, op_jop, op_li, op_limi,
	op_lwpi, op_mov, op_movb, op_mpy, op_neg, op_ori,
	op_rtwp, op_s, op_sb, op_sbo, op_sbz, op_seto, op_sla,
	op_soc, op_socb, op_sra, op_src, op_srl, op_stst,
	op_stwp, op_swpb, op_szc, op_szcb, op_tb, op_x, op_xop,
	op_xor, op_byte, op_data, op_rt, op_label, op_assy };

/*
 *      code generation structures and constants
 */


#define F_REG   1       /* data register direct mode allowed */
#define F_MEM   4       /* memory alterable modes allowed */
#define F_IMMED 8       /* immediate mode allowed */
#define F_NOIMM 7       /* all modes except for immediate F_ALL&~F_IMMED */
#define F_ALL   15      /* all modes allowed */
#define F_VOL   16      /* need volatile operand */
#define F_NOVALUE 32    /* don't need result value */

/*      addressing mode structure       */

struct amode {
        enum e_am       mode;		/* mode (e.g. am_reg */
        char            preg;		/* register, if any  */
	char		pairedreg;	/* if 1, paired register */
        char            tempflag;
        char            deep;           /* stack depth on allocation */
        struct enode    *offset;
        };

/*      output code structure   */

struct ocode {
        struct ocode    *fwd, *back;
        enum e_op       opcode;
        short           length;
        struct amode    *oper1, *oper2;
        };

/* 	intemediate assembly source structure */
struct ascode {
	struct ascode	*fwd;
	char		*stmt;		/* statement in string */
	};
