#include        <stdio.h>
#include        "c.h"
#include        "expr.h"
#include        "gen.h"
#include        "cglbdec.h"
#include        "proto.h"

/*
 *      TMS9900 C Compiler
 *
 *  Copyright 1991 by LGMA Products
 *
 *  This program is free for you to use for non-commercial purposes.
 *  It is, however, copyrighted so that commercial use is strictly
 *  forbidden.
 *
 *  Please contact:
 *
 *     LGMA PRODUCTS
 *     5618 AppleButter Hill Road
 *     Coopersburg, PA  18036
 *
 *  For suggestions, improvements, and the like.
 */

/*
 *      This module contains all of the code generation routines
 *      for evaluating expressions and conditions.
 */
/* Change History:
**
** Add new revisions to top.
**
**     Who           Date            Reason
**     ===           =========       =======================================
**     ALB            1/MAR/93       Implement fixes to %=, /= routines,
**				     as well as normal multiple/divide
**     ALB           20/Dec/91       Correct divide routine.   Allow for
**                                   unsigned divide/modulo.
**     ALB           08/Feb/92       Correct divide routine (again). Remove
**                                   swap_nodes which is only needed for 68000
*/

struct amode    *gen_deref(node,flags,size)
/*
 *      return the addressing mode of a dereferenced node.
 */
struct enode    *node;
int             flags, size;
{       struct amode    *ap1;
        struct enode    *nnode;
        int             siz1;
        switch( node->nodetype ) {		/* get load size */
                case en_v_ref:
                        siz1 = 0;
			               break;
                case en_b_ref:
                case en_ub_ref:
                        siz1 = 1;
                        break;
                case en_w_ref:
                case en_uw_ref:
                        siz1 = 2;
                        break;

                case en_l_ref:
                case en_ul_ref:
                        siz1 = 4;
                        break;
                }

        nnode = node->v.p[0];
        if( nnode->nodetype == en_add ) {
                ap1 = gen_index(nnode);
                do_extend(ap1,siz1,size,flags);
                make_legal(ap1,flags,size);
                return ap1;
                }

        else if( nnode->nodetype == en_autocon ) {
                ap1 = (struct amode *)xalloc(sizeof(struct amode));
                ap1->mode   = am_indm;
                ap1->offset = (struct enode *) 
                makenode(en_icon,(struct enode *)nnode->v.i,
                   (struct enode *)0);
                ap1->preg   = 7;
                if ( nnode->v.i < 0 ) {
                        make_legal(ap1,flags,2);
 		        if ( size == 1 ) {
		             do_extend(ap1, 2, 1, flags );
	                }
                 }else{
		        do_extend(ap1,siz1,size,flags);
                        make_legal(ap1,flags,size);
                 }
                 return ap1;
	}

/*	ap1 = gen_expr(nnode,F_REG | F_IMMED,2); */

        ap1 = gen_expr(nnode,F_REG | F_IMMED,size); /* generate address */
        if( ap1->mode == am_reg ) {
                ap1->mode = am_ind;
                do_extend(ap1,siz1,size,flags);
                make_legal(ap1,flags,size);
                return ap1;
                }
        ap1->mode = am_direct;
        do_extend(ap1,siz1,size,flags);
        make_legal(ap1,flags,size);
        return ap1;
}

int     gen_parms(plist)
/*
 *      push a list of parameters onto the stack and return the
 *      number of parameters pushed.
 *
 */
struct enode    *plist;
{       int     i;

        i = 0;
        while( plist != 0 ) {
              push_param(plist->v.p[0]);
              plist = plist->v.p[1];
              ++i;
              }
        return i;
}

struct amode    *gen_fcall(node,flags,retval)
/*
 *      generate a function call node and return the address mode
 *      of the result.
 */
struct enode    *node;
int             retval;
{       struct amode    *ap, *result;
        int             i, lab0;

/* note:  following is set up for six temporary registers (out of 16) */
        result = temp_data();
        (struct amode *) temp_data(); 
        (struct amode *) temp_data();	/* push any used data registers */

        freeop(result); freeop(result); freeop(result);

        i = gen_parms(node->v.p[1]);    /* generate parameters */
        if( node->v.p[0]->nodetype == en_nacon ) {
		gen_1code(op_bl,makeireg(12));
		gen_1code(op_data,make_limme(node->v.p[0]));
        }else{		/* passed function address, must store the
			   address dynamically in code (yuch)	*/
                ap = gen_expr(node->v.p[0],F_REG,2);
                ap->mode = am_reg;
                freeop(ap);
		lab0 = nextlabel++;
		gen_2code(op_mov,ap,make_label(lab0));
		gen_1code(op_bl,makeireg(12));
		gen_label(lab0);
		gen_1code(op_data,make_immed(0));
                }

	if ( i != 0 )
                gen_2code(op_ai,makedreg(14),make_immed(-(i*2)) );

        if ( flags & F_REG )
          result = ( struct amode * )temp_data();

        if ( retval != 0 ) {
          if( result->preg != 8 || (flags & F_REG) == 0 )
                gen_2code(op_mov,makedreg(8),result);
        }
        return result;
}

struct amode    *gen_unary(node,flags,size,op)
/*
 *      generate code to evaluate a unary minus or complement.
 */
struct enode    *node;
int             flags, size, op;
{       struct amode    *ap;
        ap = gen_expr(node->v.p[0],F_REG | F_VOL, size );
	gen_code(op,size,ap,(struct amode *)0);
        make_legal(ap,flags,size);
        return ap;
}

struct amode    *gen_asand(node,flags,size)
/*
 *      generate code to evaluate the binary and condition.
 *      this is needed on tms9900 version because the 9900 doesn't
 *      have an implicit and instruction, it must be done in two
 *      steps (inv & szc);
 */
struct enode    *node;
int             flags, size;
{       struct amode    *ap1, *ap2, *ap3;
        ap1 = gen_expr(node->v.p[0],F_VOL | F_REG,size);
	ap2 = gen_expr(node->v.p[1], F_ALL ,size);
        validate(ap1);          /* in case push occurred */

	if ( ap2->mode == am_immed ) {
	  if ( size == 1 ) {
	    ap2->offset->v.i = ap2->offset->v.i << 8;
	    gen_2code ( op_andi, ap1, ap2 );
	  }else{
            gen_code ( op_andi, size, ap1, ap2 );
	  }
        }else {
          ap3 = ( struct amode * )temp_data();
          gen_code(op_mov,size,ap2,ap3);
	  gen_1code ( op_inv, ap3 );
          validate(ap1);	/* in case push occurred */
	  gen_code  ( op_szc, size, ap3, ap1 );
          freeop(ap3);
	  }
        freeop(ap2);
        make_legal(ap1,flags,size);
        return ap1;
}

struct amode    *gen_asor(node,flags,size)
/*
 *      generate code to evaluate the binary or condition.
 */
struct enode    *node;
int             flags, size;
{       struct amode    *ap1, *ap2;
        ap1 = gen_expr(node->v.p[0],F_VOL | F_REG,size);
        ap2 = gen_expr(node->v.p[1], F_ALL ,size);
        validate(ap1);          /* in case push occurred */
        if ( ap2->mode == am_immed ) {
        if ( size == 1 ) {
            ap2->offset->v.i = ap2->offset->v.i << 8;
            gen_2code ( op_ori, ap1, ap2 );
        }else{
            gen_code ( op_ori, size, ap1, ap2 );
            }
        }else{
            gen_code  ( op_soc, size, ap2, ap1 );
        }
        freeop(ap2);
        make_legal(ap1,flags,size);
        return ap1;
}

struct amode    *gen_xbin(node,flags,size,op)
/*
 *      generate code to evaluate a restricted binary node and return 
 *      the addressing mode of the result.
 */
struct enode    *node;
int             flags, size, op;
{       struct amode    *ap1, *ap2;
        ap1 = gen_expr(node->v.p[0],F_VOL | F_REG,size);
        ap2 = gen_expr(node->v.p[1],F_REG,size);
        validate(ap1);          /* in case push occurred */
        gen_code(op,size,ap2,ap1);
        freeop(ap2);
        make_legal(ap1,flags,size);
        return ap1;
}

struct amode    *gen_modiv(node,flags,size,modflag,unsignedflag)
/*
 *      generate divide (modflag=0) and modulo (modflag=1) operators
 *	for signed (unsignedflag=0) and unsigned (unsignedflag=1)
 */
struct enode    *node;
int             flags, size, modflag,unsignedflag;
{       struct amode    *ap1, *ap2;
        int             lab0;

	ap1 = gen_expr(node->v.p[0],F_NOIMM,2);
	validate(ap1);

/* for unsigned divide, do simple divide */
	if ( unsignedflag ) {
	  ap2 = gen_expr ( node->v.p[1],F_NOIMM,2);
	  validate(ap2);
	  gen_2code(op_mov,ap1,makedreg(10));
	  gen_1code(op_clr,makedreg(9));
	  gen_2code(op_div,ap2,makedreg(9));
	  freeop(ap1);
	  freeop(ap2);
	  if ( modflag ) {
	    ap1 = makedreg(10);
	  }else{
	    ap1 = makedreg(9);
	  }
/* for signed divide, xor the two inputs to single register for later */
	}else{
	  ap2 = gen_expr ( node->v.p[1], F_REG, 2);
	  validate(ap2);
	  gen_2code(op_mov,ap2,makedreg(0));
	  gen_2code(op_mov,ap1,makedreg(10));
	  gen_2code(op_xor,makedreg(10), makedreg(0));
	  gen_1code(op_abs,ap2);
	  gen_1code(op_abs,makedreg(10));
	  gen_1code(op_clr,makedreg(9));
	  gen_2code(op_div,ap2,makedreg(9));
          gen_2code ( op_andi, makedreg(0), make_immed( 0x8000 ) );
	  freeop(ap1);
	  freeop(ap2);

          lab0 = nextlabel++;
          gen_1code ( op_jeq,  make_labei(lab0));
          if ( modflag ) {
            gen_1code ( op_neg, makedreg(10));
            gen_label ( lab0 );
	    ap1 = makedreg(10);
          }else {
            gen_1code ( op_neg, makedreg(9) );
            gen_label ( lab0 );
	    ap1 = makedreg(9);
          }
	}
	do_extend(ap1,2,size);
        return ap1;
}

struct amode    *gen_mul(node,flags,size)
/*
 *      generate code to evaluate a multiply node. 
 */
struct enode    *node;
int             flags, size;
{       struct amode    *ap1, *ap2;

        ap1 = gen_expr(node->v.p[0],F_NOIMM,2);
        ap2 = gen_expr(node->v.p[1],F_NOIMM,2);
	validate(ap1);
	validate(ap2);
        gen_loadr(ap2,size,9);
        gen_2code(op_mpy,ap1,makedreg(9));
	freeop(ap2);
	freeop(ap1);
	ap1 = makedreg(10);
        return ap1;
}

struct amode    *gen_asmul(node,flags,size)
/*
 *      generate code to evaluate a *= multiply node. 
 */
struct enode    *node;
int             flags, size;
{       struct amode    *ap1, *ap2;

        ap1 = gen_expr(node->v.p[0],F_NOIMM,2);
        ap2 = gen_expr(node->v.p[1],F_NOIMM,2);
	validate(ap1);
	validate(ap2);
        gen_loadr(ap2,size,9);
        gen_2code(op_mpy,ap1,makedreg(9));
	freeop(ap2);
	freeop(ap1);
	make_legal(ap1,flags,size);
	gen_2code(op_mov,makedreg(10),ap1);
}

struct amode    *gen_hook(node,flags,size)
/*
 *      generate code to evaluate a condition operator node (?:)
 */
struct enode    *node;
int             flags, size;
{       struct amode    *ap1, *ap2;
        int             false_label, end_label;
        false_label = nextlabel++;
        end_label = nextlabel++;
        flags = (flags &  F_REG ) | F_VOL;
        falsejp(node->v.p[0],false_label);
        node = node->v.p[1];
        ap1 = gen_expr(node->v.p[0],flags,size);
        freeop(ap1);
        gen_1code(op_b,make_label(end_label));
        gen_label(false_label);
        ap2 = gen_expr(node->v.p[1],flags,size);
        if( !equal_address(ap1,ap2) ) {
                freeop(ap2);
                (struct amode *) temp_data();
                gen_code(op_mov,size,ap2,ap1);
                }
        gen_label(end_label);
        return ap1;
}

struct amode    *gen_asadd(node,flags,size,op)
/*
 *      generate a plus equal +=  or a minus -= equal node.
 */
struct enode    *node;
int             flags,size,op;
{       struct amode    *ap1, *ap2;
        int             ssize;
        ssize = natural_size(node->v.p[0]);
        if( ssize > size )
                size = ssize;
        ap1 = gen_expr(node->v.p[0],F_ALL,ssize);
        ap2 = gen_expr(node->v.p[1], F_REG | F_IMMED,size);
        validate(ap1);
        gen_load ( ap2, ssize );
        gen_code ( op, ssize, makedreg(8), ap1);
        freeop(ap2);
        do_extend(ap1,ssize,size);
        make_legal(ap1,flags,size);
        return ap1;
}

struct amode   *gen_asshift(node,flags,size,op)
/*
 *      generate shift equals operators.
 */
struct enode    *node;
int             flags, size, op;
{       struct amode    *ap1, *ap2, *ap3;
        ap1 = gen_expr(node->v.p[0],F_ALL,size);
        if( ap1->mode != am_reg ) {
                ap3 = ( struct amode * )temp_data();
                gen_code(op_mov,size,ap1,ap3);
                }
        else
                ap3 = ap1;
        ap2 = gen_expr(node->v.p[1],F_REG | F_IMMED,size);
        validate(ap3);
        gen_code(op,size,ap3,ap2);
        freeop(ap2);
        if( ap3 != ap1 ) {
                gen_code(op_mov,size,ap1,ap3);
                freeop(ap3);
                }
        make_legal(ap1,flags,size);
        return ap1;
}

struct amode    *gen_asmodiv(node,flags,size,modflag)
/*
 *      generate /= and %= nodes.
 */
struct enode    *node;
int             flags, size, modflag;
{       struct amode    *ap1, *ap2;
        int             lab0;

        ap1 = gen_expr(node->v.p[0],F_NOIMM,2);
        ap2 = gen_expr(node->v.p[1],F_NOIMM,2);

	gen_2code(op_mov,ap2,makedreg(0));
	gen_2code(op_mov,ap1,makedreg(10));
	gen_2code(op_xor,makedreg(10), makedreg(0));
	gen_1code(op_abs,makedreg(10));
	gen_1code(op_clr,makedreg(9));
	gen_2code(op_div,ap2,makedreg(9));
        gen_2code ( op_andi, makedreg(0), make_immed( 0x8000 ) );

        lab0 = nextlabel++;
        gen_1code ( op_jeq,  make_labei(lab0));
        if ( modflag ) {
          gen_1code ( op_neg, makedreg(10));
          gen_label ( lab0 );
          gen_2code ( op_mov, makedreg(10), ap1 );
        }else {
          gen_1code ( op_neg, makedreg(9) );
          gen_label ( lab0 );
          gen_2code ( op_mov, makedreg(9), ap1 );
        }
        make_legal(ap1,flags,2);
        freeop(ap2);
        do_extend(ap1,2,size);
        return ap1;
}

struct amode    *gen_assign(node,flags,size)
/*
 *      generate code for an assignment node. if the size of the
 *      assignment destination is larger than the size passed then
 *      everything below this node will be evaluated with the
 *      assignment size.
 */
struct enode    *node;
int             flags, size;
{       struct amode    *ap1, *ap2;
        int             ssize;
        switch( node->v.p[0]->nodetype )
                {
		case en_v_ref:
			ssize = 0;
			break;
                case en_b_ref:
                case en_ub_ref:
                        ssize = 1;
                        break;
                case en_w_ref:
		case en_tempref:
                case en_uw_ref:
                        ssize = 2;
                        break;
                case en_l_ref:
                case en_ul_ref:
                        ssize = 4;
                        break;
                }
        ap2 = gen_expr(node->v.p[1],F_ALL,size);
        ap1 = gen_expr(node->v.p[0],F_ALL,ssize);
        validate(ap2);

/* if input/output not same size, then do extend either way */
        if ( ssize != size ) {
          gen_load  ( ap2, size );
	  if ( size == 2 ) {
             gen_2code ( op_sla, makedreg(8), make_immed(8));
	  }else{
             gen_2code ( op_sra, makedreg(8), make_immed(8));
	  }
	  gen_code  ( op_mov, ssize, makedreg(8), ap1);

/* input, output same size.  If immediate, do load thru r8.  If not
   immediate, then just move it into place                           */

	}else if ( ap2->mode == am_immed ) {
          if ( ap1->mode == am_reg ) {
            if ( ssize == 1 ) {
              gen_code ( op_li, ssize, make_immed(ap1->offset->v.i<<8),ap2);
            }else{
              gen_code ( op_li, ssize, ap1, ap2);
            }
          }else{
            gen_load ( ap2, ssize );
            gen_code(op_mov,ssize,makedreg(8),ap1);
          }
	}else{
          gen_code(op_mov,ssize,ap2,ap1);
	}
        freeop(ap1);
        return ap2;
}

struct amode    *gen_aincdec(node,flags,size,op)
/*
 *      generate an auto increment or decrement node. op should be
 *      either op_a (for increment) or op_s (for decrement).
 */
struct enode    *node;
int             flags, size;
{       struct amode    *ap1, *ap2;
        int             siz1;
        siz1 = natural_size(node->v.p[0]);
        if( flags & F_NOVALUE )         /* dont need result */
                {
                ap1 = gen_expr(node->v.p[0],F_ALL,siz1);
                gen_load ( make_immed(node->v.p[1]), siz1 );
		gen_code ( op, siz1, makedreg(8), ap1);
                freeop(ap1);
                return ap1;
                }
        ap1 = (struct amode *) temp_data();
        ap2 = gen_expr(node->v.p[0],F_ALL,siz1);
        validate(ap1);
        gen_code(op_mov,siz1,ap2,ap1);
	gen_load(make_immed(node->v.p[1]), siz1);
	gen_code(op, siz1, makedreg(8), ap2);
        freeop(ap2);
        do_extend(ap1,siz1,size);
        return ap1;
}

struct amode    *gen_expr(node,flags,size)
/*
 *      general expression evaluation. returns the addressing mode
 *      of the result.
 */
struct enode    *node;
int             flags, size;
{       struct amode    *ap1;
        int             lab0, lab1;
        int             natsize;
        if( node == 0 ) {
                printf("DIAG - null node in gen_expr.\n");
                return 0;
                }
        switch( node->nodetype ) {
                case en_icon:
                case en_labcon:
                case en_nacon:
                        ap1 = (struct amode*)xalloc(sizeof(struct amode));
                        ap1->mode = am_immed;
                        ap1->offset = node;
                        make_legal(ap1,flags,size);
                        return ap1;

                case en_autocon:
			if ( node->v.i < 0 ) {
                             ap1 = (struct amode*)xalloc(sizeof(struct amode));
                             ap1->mode = am_indm;
                             ap1->preg = 7;
                             ap1->offset = node;
                             make_legal(ap1,flags,size);
                        }else{
                             ap1 = temp_data();	/* allocate temp register */
    	                     gen_2code (op_mov, makedreg(7), ap1 );
	                     if ( node->v.i != 0 ) {
	                       gen_2code (op_ai,ap1,make_immed(node->v.i));
	                     }
                             make_legal(ap1,flags,size);
			}
                        return ap1;

		case en_v_ref:
                case en_b_ref:
                case en_w_ref:
                case en_l_ref:
                case en_ub_ref:
                case en_uw_ref:
                case en_ul_ref:
                        return gen_deref(node,flags,size);
                case en_tempref:
                        ap1 = (struct amode*)xalloc(sizeof(struct amode));
                        ap1->mode = am_reg;
                        ap1->preg = node->v.i;
                        ap1->tempflag = 0;      /* not a temporary */
                        if ( size != 2 )
                          {
                          gen_2code(op_mov, ap1, makedreg(11));
                          ap1->preg = 11;
                          do_extend(ap1,2,size,flags);
			  }
                        make_legal(ap1,flags,size);
                        return ap1;
/* note : sizes on inv & neg forced to 2, since will work for 1 */
                case en_uminus:
                        return gen_unary(node,flags,size,op_neg);
                case en_compl:
                        return gen_unary(node,flags,size,op_inv);
                case en_add: 
                        return gen_binary(node,flags,size,op_a);
                case en_sub:
                        return gen_binary(node,flags,size,op_s);
                case en_and:
			return gen_asand ( node, flags, size);
                case en_or:
                        return gen_asor(node,flags,size);
		case en_xor:
			return gen_xbin(node,flags,size,op_xor);
                case en_mul:
                        return gen_mul(node,flags,size);
                case en_umul:
                        return gen_mul(node,flags,size);
                case en_div:
                        return gen_modiv(node,flags,size,0,0);
                case en_udiv:
                        return gen_modiv(node,flags,size,0,1);
                case en_mod:
                        return gen_modiv(node,flags,size,1,0);
                case en_umod:
                        return gen_modiv(node,flags,size,1,1);
                case en_lsh:
                        return gen_shift(node,flags,size,op_sla);
                case en_rsh:
                        return gen_shift(node,flags,size,op_sra);
                case en_rshl:
                        return gen_shift(node,flags,size,op_srl);
                case en_asadd:
                        return gen_asadd(node,flags,size,op_a);
                case en_assub:
                        return gen_asadd(node,flags,size,op_s);
                case en_asand:
			return gen_asand ( node, flags, size );
                case en_asor:
                        return gen_asor  ( node, flags, size);
                case en_aslsh:
                        return gen_asshift(node,flags,size,op_sla);
                case en_asrsh:
                        return gen_asshift(node,flags,size,op_sra);
                case en_asmul:
			return gen_asmul(node,flags,size);
                case en_asdiv:
                        return gen_asmodiv(node,flags,size,0);
                case en_asmod:
                        return gen_asmodiv(node,flags,size,1);
                case en_assign:
                        return gen_assign(node,flags,size);
                case en_ainc:
                        return gen_aincdec(node,flags,size,op_a);
                case en_adec:
                        return gen_aincdec(node,flags,size,op_s);
                case en_land:   case en_lor:
                case en_eq:     case en_ne:
                case en_lt:     case en_le:
                case en_gt:     case en_ge:
                case en_ult:    case en_ule:
                case en_ugt:    case en_uge:
                case en_not:
                        lab0 = nextlabel++;
                        lab1 = nextlabel++;
                        falsejp(node,lab0);
                        ap1 = (struct amode *)temp_data();

			if ( size == 1 )
			  gen_2code(op_li,ap1,make_immed(1<<8));
			else
			  gen_2code(op_li,ap1,make_immed(1));

                        gen_1code(op_b,make_label(lab1));
                        gen_label(lab0);
                        gen_1code(op_clr,ap1);
                        gen_label(lab1);
                        return ap1;
                case en_cond:
                        return gen_hook(node,flags,size);
                case en_void:
                     natsize = natural_size(node->v.p[0]);
                     freeop(gen_expr(node->v.p[0],F_ALL | F_NOVALUE,natsize));
                     return gen_expr(node->v.p[1],flags,size);
                case en_fcall:
                        ap1 = gen_fcall(node,flags,1);
			do_extend(ap1, 2, size, flags);
			return ap1;
                case en_vfcall:
                        ap1 = gen_fcall(node,flags,0);
                        return ap1;
                default:
                        printf("DIAG - uncoded node in gen_expr.\n");
                        return 0;
                }
}

void gen_compare(node)
/*
 *      generate code to do a comparison of the two operands of
 *      node.
 */
struct enode    *node;
{       struct amode    *ap1, *ap2;
        int             size;
        size = natural_size(node);

        ap2 = gen_expr(node->v.p[1],F_ALL,size);
	if ( ap2->mode == am_immed )
	  ap1 = gen_expr(node->v.p[0],F_REG,size);
	else
          ap1 = gen_expr(node->v.p[0],F_ALL, size);

        validate(ap2);				/* was ap1 */
	validate(ap1);
	if ( ap2->mode == am_immed ) {
	  if ( size == 1 ) {
	    gen_load ( ap2, size );
	    gen_code ( op_c, size, ap1, makedreg(8));
          }
	  else{
	    gen_code ( op_ci, size, ap1, ap2 );
          }
        }else{
            if ( ap1->mode == am_immed ) {
              gen_load ( ap1, size );
              gen_code ( op_c,   size, makedreg(8), ap2 );
            }else{
	      gen_code ( op_c,  size, ap1, ap2 );
            }
	}
        freeop(ap2);
        freeop(ap1);
}

void truejp(node,label)
/*
 *      generate a jump to label if the node passed evaluates to
 *      a true condition.
 */
struct enode    *node;
int             label;
{       struct amode    *ap1;
        int             siz1;
        int             lab0;
        if( node == 0 )
                return;
        switch( node->nodetype ) {
                case en_eq:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jne,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_ne:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jeq,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_lt:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jgt,make_labei(lab0));
			gen_1code(op_jeq,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_le:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jgt,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
			break;
                case en_gt:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jlt,make_labei(lab0));
			gen_1code(op_jeq,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_ge:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jlt,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_ult:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jhe,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_ule:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jh,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_ugt:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jle,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_uge:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jl,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_land:
                        lab0 = nextlabel++;
                        falsejp(node->v.p[0],lab0);
                        truejp(node->v.p[1],label);
                        gen_label(lab0);
                        break;
                case en_lor:
                        truejp(node->v.p[0],label);
                        truejp(node->v.p[1],label);
                        break;
                case en_not:
                        falsejp(node->v.p[0],label);
                        break;
                default:
                        siz1 = natural_size(node);
                        ap1 = gen_expr(node,F_ALL,siz1);
                        gen_load(ap1,siz1);
                        freeop(ap1);
			lab0 = nextlabel++;
			gen_1code(op_jeq,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                }
}

void falsejp(node,label)
/*
 *      generate code to execute a jump to label if the expression
 *      passed is false.
 */
struct enode    *node;
int             label;
{       struct amode    *ap;
        int             siz1;
        int             lab0;
        if( node == 0 )
                return;
        switch( node->nodetype ) {
                case en_eq:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jeq,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_ne:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jne,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_lt:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jlt,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_le:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jlt,make_labei(lab0));
			gen_1code(op_jeq,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_gt:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jgt,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_ge:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jgt,make_labei(lab0));
			gen_1code(op_jeq,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_ult:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jl,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_ule:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jle,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_ugt:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jh,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_uge:
                        gen_compare(node);
			lab0 = nextlabel++;
			gen_1code(op_jhe,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                case en_land:
                        falsejp(node->v.p[0],label);
                        falsejp(node->v.p[1],label);
                        break;
                case en_lor:
                        lab0 = nextlabel++;
                        truejp(node->v.p[0],lab0);
                        falsejp(node->v.p[1],label);
                        gen_label(lab0);
                        break;
                case en_not:
                        truejp(node->v.p[0],label);
                        break;
                default:
                        siz1 = natural_size(node);
                        ap = gen_expr(node,F_ALL,siz1);
			gen_load(ap,siz1);
                        freeop(ap);
			lab0 = nextlabel++;
			gen_1code(op_jne,make_labei(lab0));
			gen_1code(op_b,make_label(label));
			gen_label(lab0);
                        break;
                }
}
