#include        <stdio.h>
#include        "c.h"
#include        "expr.h"
#include        "gen.h"
#include        "cglbdec.h"
#include        "proto.h"

/*
 *      TMS9900 C Compiler
 *
 *  Copyright 1990 by LGMA Products
 *
 *  This program is free for you to use for non-commercial purposes.
 *  It is, however, copyrighted so that commercial use is strictly
 *  forbidden.
 *
 *  Please contact:
 *
 *     LGMA PRODUCTS
 *     5618 AppleButter Hill Road
 *     Coopersburg, PA  18036
 *
 *  For suggestions, improvements, and the like.
 *
 */

void genxasm(stmt)
/*
 *      generate the code to evaluate the #asm statement
*/
struct snode    *stmt;
{       

	if ( stmt->s1 != 0 ) {	/* if assembly code attached */
	   	gen_1code ( op_assy, makeassy(stmt->s1) );
	}
}

void genwhile(stmt)
/*
 *      generate code to evaluate a while statement.
 */
struct snode    *stmt;
{       int     lab1, lab2;
        initstack();            /* initialize temp registers */
        lab1 = contlab;         /* save old continue label */
        lab2 = breaklab;        /* save old break label */
        contlab = nextlabel++;  /* new continue label */
        gen_label(contlab);
        if( stmt->s1 != 0 )      /* has block */
                {
                breaklab = nextlabel++;
                initstack();
                falsejp(stmt->exp,breaklab);
                genstmt(stmt->s1);
                gen_1code(op_b,make_label(contlab));
                gen_label(breaklab);
                breaklab = lab2;        /* restore old break label */
                }
        else					        /* no loop code */
                {
                initstack();
                truejp(stmt->exp,contlab);
                }
        contlab = lab1;         /* restore old continue label */
}

void gen_for(stmt)
/*
 *      generate code to evaluate a for loop
 */
struct snode    *stmt;
{       int     old_break, old_cont, exit_label, loop_label;
        old_break = breaklab;
        old_cont = contlab;
        loop_label = nextlabel++;
        exit_label = nextlabel++;
        contlab    = nextlabel++;		/* new */
        initstack();
        if( stmt->label != 0 )
                gen_expr((struct enode *)(stmt->label),F_ALL | F_NOVALUE
                        ,natural_size((struct enode *)stmt->label));
        gen_label(loop_label);
        initstack();
        if( stmt->exp != 0 )
                falsejp(stmt->exp,exit_label);
        if( stmt->s1 != 0 )
		{
                breaklab = exit_label;
                genstmt(stmt->s1);
		}
        initstack();
        gen_label(contlab);
        if( stmt->s2 != 0 )
                gen_expr((struct enode *)stmt->s2,F_ALL | F_NOVALUE,
			natural_size((struct enode *)stmt->s2));
        gen_1code(op_b,make_label(loop_label));
        breaklab = old_break;
        contlab = old_cont;
        gen_label(exit_label);
}

void genif(stmt)
/*
 *      generate code to evaluate an if statement.
 */
struct snode    *stmt;
{       int     lab1, lab2;
        lab1 = nextlabel++;     /* else label */
        lab2 = nextlabel++;     /* exit label */
	initstack();            /* clear temps */
        falsejp(stmt->exp,lab1);
	genstmt(stmt->s1);
	if( stmt->s2 != 0 )             /* else part exists */
		{
		gen_1code(op_b,make_label(lab2));
		gen_label(lab1);
		genstmt(stmt->s2);
		gen_label(lab2);
		}
	else                            /* no else code */
		gen_label(lab1);
}

void gendo(stmt)
/*
 *      generate code for a do - while loop.
 */
struct  snode    *stmt;
{       int     oldcont, oldbreak;
        oldcont = contlab;
        oldbreak = breaklab;
        contlab = nextlabel++;
        gen_label(contlab);
        if( stmt->s1 != 0 && stmt->s1->next != 0 )
                {
                breaklab = nextlabel++;
                genstmt(stmt->s1);      /* generate body */
                initstack();
                truejp(stmt->exp,contlab);
                gen_label(breaklab);
                }
        else
                {
                genstmt(stmt->s1);
                initstack();
                truejp(stmt->exp,contlab);
                }
        breaklab = oldbreak;
        contlab = oldcont;
}

void call_library(lib_name)
/*
 *      generate a call to a library routine.
 */
char    *lib_name;
{       SYM     *sp;
        sp = (SYM *)gsearch(lib_name);
        if( sp == 0 )
                {
                ++global_flag;
                sp = (SYM *)xalloc(sizeof(SYM));
                sp->tp = &stdfunc;
                sp->name = lib_name;
                sp->storage_class = sc_external;
                insert(sp,&gsyms);
                --global_flag;
                }
        gen_1code(op_bl,make_strlab(lib_name));
}

void genswitch(stmt)
/*
 *      generate a linear search switch statement.
 */
struct snode    *stmt;
{       int             curlab;
        struct snode    *defcase;
        struct amode    *ap;
        curlab = nextlabel++;
        defcase = 0;
        initstack();
        ap = (struct amode *)gen_expr(stmt->exp,F_REG | F_VOL,2);
        if( ap->preg != 0 )
                gen_load( ap, 2 );
        stmt = stmt->s1;

	gen_1code ( op_bl, makeireg(15));

        while( stmt != 0 )
                {
                if( stmt->s2 )          /* default case ? */
                        {
                        stmt->label = (int *)curlab;
                        defcase = stmt;
                        }
                else
                        {
                        gen_2code(op_data,make_labei(curlab),
                                make_immed((struct enode *)stmt->label));
                        stmt->label = (int *)curlab;
                        }
                if( stmt->s1 != 0 && stmt->next != 0 )
                        curlab = nextlabel++;
                stmt = stmt->next;
                }
	gen_1code(op_data,make_immed(0));
        if( defcase == 0 )
		gen_1code(op_b,make_label(breaklab));
	else
		gen_1code(op_b,make_label(defcase->label));
}

void gencase(stmt)
/*
 *      generate all cases for a switch statement.
 */
struct snode    *stmt;
{       while( stmt != 0 )
                {
                if( stmt->s1 != 0 )
                        {
                        gen_label((int)stmt->label);
                        genstmt(stmt->s1);
                        }
                else if( stmt->next == 0 )
                        gen_label((int)stmt->label);
                stmt = stmt->next;
                }
}

void genxswitch(stmt)
/*
 *      analyze and generate best switch statement.
 */
struct snode    *stmt;
{       int     oldbreak;
        oldbreak = breaklab;
        breaklab = nextlabel++;
        genswitch(stmt);
        gencase(stmt->s1);
        gen_label(breaklab);
        breaklab = oldbreak;
}

void genreturn(stmt,autovar)
/*
 *      generate a return statement.
 */
struct snode    *stmt;
int    autovar;
{       struct amode    *ap;

        if( stmt != 0 && stmt->exp != 0 )
                {
                initstack();
                ap = (struct amode *) gen_expr(stmt->exp,F_ALL,2);
                if( ap->mode != am_reg || ap->preg != 0 )
                        gen_load(ap,2);
                }
        if( retlab == -1 )
                {
                retlab = nextlabel++;
                gen_label(retlab);
		gen_1code(op_b,makeireg(13));
                }
        else
                gen_1code(op_b,make_label(retlab));
}

void genstmt(stmt)
/*
 *      genstmt will generate a statement and follow the next pointer
 *      until the block is generated.
 */
struct snode    *stmt;
{       
	while( stmt != 0 )
                {
                switch( stmt->stype )
                        {
                        case st_label:
                                gen_label((int)stmt->label);
                                break;
                        case st_goto:
                                gen_1code(op_b,make_label(stmt->label));
                                break;
                        case st_expr:
                                initstack();
                                gen_expr(stmt->exp,F_ALL | F_NOVALUE,
                                        natural_size(stmt->exp));
                                break;
                        case st_return:
                                genreturn(stmt);
                                break;
                        case st_if:
                                genif(stmt);
                                break;
                        case st_while:
                                genwhile(stmt);
                                break;
                        case st_for:
                                gen_for(stmt);
                                break;
			case st_do:
				gendo(stmt);
				break;
                        case st_continue:
                                gen_1code(op_b,make_label(contlab));
                                break;
                        case st_break:
                                gen_1code(op_b,make_label(breaklab));
                                break;
                        case st_switch:
                                genxswitch(stmt);
                                break;
			case st_asm:
				genxasm(stmt);
				break;
                        default:
                                printf("DIAG - unknown statement.\n");
                                break;
                        }
                stmt = stmt->next;
                }
}

void genfunc(stmt)
/*
 *      generate a function body.
 */
struct snode    *stmt;
{       retlab = contlab = breaklab = -1;
	if( lc_auto & 1 )			/* if frame size odd */
		++lc_auto;			/* make it even */

/* now push stack pointer to start of local data area                 */

	if ( lc_auto != 0 ) {
	    gen_2code(op_ai,makedreg(14),make_immed(lc_auto));
	}
        opt1(stmt);			/* execute statement optimization */
        genstmt(stmt,lc_auto);
        genreturn((struct snode *)0);
}

