#include        <stdio.h>
#include        "c.h"
#include        "expr.h"
#include        "gen.h"
#include        "cglbdec.h"
#include        "proto.h"

/*
 *      TMS9900 C Compiler
 *
 *  Copyright 1990 by LGMA Products
 *  This program is free for you to use for non-commercial purposes.
 *  It is, however, copyrighted so that commercial use is strictly
 *  forbidden.
 *
 *  Please contact:
 *
 *     LGMA PRODUCTS
 *     5618 AppleButter Hill Road
 *     Coopersburg, PA  18036
 *
 *  For suggestions, improvements, and the like.
** Change History:
**
** Add new revisions to top.
**
**     Who           Date            Reason
**     ===           =========       =========================================
**     ALB           20/Dec/91       Add char forms of \v & \a per ANSI
**                                   recommendations & unsigned constants.
**
 *  getsym contains the routines for:
 *
 *   initsym  -  initialization of various flags/pointers for compile
 *   get_line  -  display last line processed + errors, get next line
 *   getch    -  get next character from input line
 *   error    -  buffer an error on input line
 *   getid    -  get an identifier from input line
 *   getsch   -  get quoted string
 *   radix36  -  get a radix 36 number
 *   getbase  -  get an integer in any base
 *   getfrac  -  get a fractional part of number
 *   getexp   -  get exponent
 *   getnum   -  get a number
 *   getsym   -  get a symbol from input line
 *   needpunc -  checks any needed punctuation

 */

/*  initsym -  setup for processing, initialize a number of symbols
               when the TIC compiler is first entered              */
void initsym()
{       lptr = inpline;			/* line pointer */
        inpline[0] = 0;			/* null terminate the line */
        numerrs = 0;			/* reset # of errors */
        total_errors = 0;		/* reset total error count */
        lineno = 0;			/* reset current line number */
	ifdepth = -1;			/* depth of #ifdef's */
}

/*  get_line -  prints the previous line to the print file, and if
               any errors have been buffered for this line, displays
               each error in the print file and on the screen.
               Retrieve the next line from the file and place in the
               "inpline" char array                                    */
int     get_line()
{       int     rv, errnumb;
        static char  *errtxt[] = {  "Syntax",
                             "Illegal Character",
                             "Bad Floating Point Constant",
                             "Illegal Type",
                             "Undefined Symbol",
                             "Duplicate Symbol",
                             "Punctuation",
                             "Identifier Expected",
                             "Not Initialized",
                             "Incomplete",
                             "Illegal Initialization",
                             "Illegal Initialization Size",
                             "Illegal Class",
                             "Bad Block",
                             "No Pointer",
                             "No Function",
                             "No Member",
                             "Bad Last Value",
                             "Bad Type in Dereference Operation",
                             "MisMatch",
                             "Expression Expected",
                             "While Expected",
                             "Missing CASE Statement",
                             "Duplicate CASE Statement",
                             "Bad Label",
                             "Bad Pre-Processor",
                             "on Include File",
                             "cannot open file",
                             "Bad #define",
                             "not expecting #endif or #else",
                             "too many #ifdef's",
                             "unknown error"    };

/* do until valid statement processed               */
start:
   for ( ;; ) {
        if( lineno > 0 ) {
                if ( (incldepth == 0) && listflag ) {
		  fprintf(list, "%s", inpline);
		}
/* if errors encountered, display the line & process each error */
                if ( numerrs != 0 ) {
                  printf("\n%d ",lineno);
                  printf("%s", inpline);
                  while(numerrs--) {
                    errnumb = errnos[numerrs];
                    if ( errnumb < 0 || 
                         errnumb >= ERR_UNKNOWN ) {
                       errnumb = ERR_UNKNOWN;
			 }
                    printf("***Error %d, ",errnumb);
                    printf("%s", errtxt[errnumb] );
                    printf("\n");
                    if ( (incldepth == 0)  && listflag ) {
                      fprintf(list,"***Error %d, ",errnumb);
                      fprintf(list, "%s", errtxt[errnumb] );
                      fprintf(list, "\n");
		    }
		  }
		}
                numerrs = 0;
	}
        ++lineno;
        if ( !quiet ) {
          printf("Line: %d\r",lineno);
	}
        rv = (fgets(inpline,131,input) == NULL);
        if( rv && incldepth > 0 ) {
                fclose(input);
                input  = inclfile[--incldepth];
                lineno = inclline[incldepth];
                goto start;
                }
        if( rv ) {
                return TRUE;
	}
        lptr = inpline;
/* if processing #asm, then just return to pre=processor */
	if ( assembly ) {		/* process assembly statement */
	  return FALSE;
	}
/* if not processing #ifdef, or if processing #ifdef && process mode is TRUE */
        if ( ifdepth < 0 || ifarray[ifdepth] == TRUE ) {
          if(inpline[0] == '#')  {
	       if ( preprocess() ) {
		 goto start;
	       }else{
		 return FALSE;		/* process #asm or #endasm */
	       }
          }else{
               return FALSE;
          }
	}
/* processing #ifdef, and process mode is false.  Skip all entries until
   another preprocess command is found */
        if (inpline[0] == '#' ) {
          preprocess();
          goto start;
   	}
   }
}
/*
 *      getch - basic get character routine.
 */
void     getch()
{       while( (lastch = (int) *lptr++) == '\0') {

                if( lstackptr > 0 ) {
                        lptr = linstack[--lstackptr];
                        lastch = (int) chstack[lstackptr];
                        return;
                        }

                if ( get_line() != 0 ) {
                  lastch = -1;
                  return;
		}
        }
        return;
}

/*
 *      error - buffer error information
 */
void error(n)
int     n;
{       if ( numerrs < MAX_ERROR ) {
	  errnos[numerrs++] = n;
        }
        ++total_errors;
}

/*
 *      getid - get an identifier.
 *
 *      identifiers are any isidch conglomerate
 *      that doesn't start with a numeric character.
 *      this set INCLUDES keywords.
 */
int     getid()
{       register int    i;
        i = 0;
        while(isidch(lastch)) {
                if(i < 19) {
                        lastid[i++] = lastch;
		}
                getch();
                }
        lastid[i] = '\0';
        lastst = id;
}
 
/*
 *      getsch - get a character in a quoted string.
 *
 *      this routine handles all of the escape mechanisms
 *      for characters in strings and character constants.
 */
int     getsch()        /* return an in-quote character */
{       register int    i, j;
        if(lastch == '\n')
                return -1;
        if(lastch != '\\') {
                i = lastch;
                getch();
                return i;
                }
        getch();        /* get an escaped character */
        if(isdigit(lastch)) {
                i = 0;
                for(j = 0;j < 3;++j) {
                        if(lastch <= '7' && lastch >= '0')
                                i = (i << 3) + lastch - '0';
                        else
                                break;
                        getch();
                        }
                return i;
                }
        i = lastch;
        getch();
        switch(i) {
                case '\n':
                        getch();
                        return getsch();
                case 'b':
                        return '\b';
                case 'f':
                        return '\f';
                case 'n':
                        return '\n';
                case 'r':
                        return '\r';
		case 't':
			return '\t';
                case 'v':
                        return '\v';
                case 'a':
                        return '\a';
                default:
                        return i;
                }
}

int     radix36(c)
char    c;
{       if(isdigit(c))
                return c - '0';
        if(c >= 'a' && c <= 'z')
                return c - 'a' + 10;
        if(c >= 'A' && c <= 'Z')
                return c - 'A' + 10;
        return -1;
}
 
/*
 *      getbase - get an integer in any base.
 */
void getbase(b)
int b;
{       register int    i, j;
        i = 0;
        while(isalnum(lastch)) {
                if((j = radix36(lastch)) < b) {
                        i = i * b + j;
                        getch();
                        }
                else break;
                }
        ival = i;
        lastst = iconst;
}
 
/*
 *      getfrac - get fraction part of a floating number.
 */
void getfrac()
{
/*        double  frmul;
        frmul = 0.1;
        while(isdigit(lastch)) {
                rval += frmul * (lastch - '0');
                getch();
                frmul *= 0.1;
                }				*/
}
 
/*
 *      getexp - get exponent part of floating number.
 *
 *      this algorithm is primative but usefull.  Floating
 *      exponents are limited to +/-255 but most hardware
 *      won't support more anyway.
 */
void getexp()
{
/*      double  expo, exmul;
        expo = 1.0;
        if(lastst != rconst)
                rval = (0.0) + ival;
        if(lastch = '-') {
                exmul = 0.1;
                getch();
                }
        else
                exmul = 10.0;
        getbase(10);
        if(ival > 255)
                error(ERR_FPCON);
        else
                while(ival--)
                        expo *= exmul;
        rval *= expo;				*/
}
 
/*
 *      getnum - get a number from input.
 *
 *      getnum handles all of the numeric input. it accepts
 *      decimal, octal, hexidecimal, and floating point numbers.
 */
void getnum()
{       register int    i;
        i = 0;
        if(lastch == '0') {
                getch();
                if(lastch == 'x' || lastch == 'X') {
                        getch();
                        getbase(16);
                        }
                else getbase(8);
                }
        else    {
                getbase(10);
                if(lastch == '.') {
                        getch();
/*                        rval = (0.0) + ival;    /* float the integer part */
                        getfrac();      /* add the fractional part */
                        lastst = rconst;
                        }
                if(lastch == 'e' || lastch == 'E') {
                        getch();
                        getexp();       /* get the exponent */
                        }
						/* handle unsigned const */
                if(lastch == 'U' || lastch == 'u') {
                        lastst = uiconst;
                        getch();
		}
	}
}
 
/*
 *      getsym - get next symbol from input stream.
 *
 *      getsym is the basic lexical analyzer.  It builds
 *      basic tokens out of the characters on the input
 *      stream and sets the following global variables:
 *
 *      lastch:         A look behind buffer.
 *      lastst:         type of last symbol read.
 *      laststr:        last string constant read.
 *      lastid:         last identifier read.
 *      ival:           last integer constant read.
 *      rval:           last real constant read.
 *
 *      getsym should be called for all your input needs...
 */
void    getsym()
{       register int    i, j;
        SYM             *sp;
        char            llastch;

restart:        /* we come back here after comments */
	if ( assembly ) {
		lastst = kw_asm;
		return;
	}
	if ( lastch == ' ' || lastch == '\t' || lastch == '\n' ) {
		getch();
		goto restart;
	}
        if( lastch == -1) {
                lastst = kw_eof;
		return;
	}
        else if(isdigit(lastch))  {
                getnum();
	}
        else if(isidch(lastch)) {
                getid();
                if( (sp = (SYM *) search(lastid,defsyms.head)) != 0 ) {
                        linstack[lstackptr] = lptr;
                        chstack[lstackptr++] = lastch;
                        lptr = sp->value.s;
                        getch();
                        goto restart;
                        }
        }else{
                llastch = lastch;
                getch();
                switch(llastch) {
                case '+':
                        if(lastch == '+') {
                                getch();
                                lastst = autoinc;
                                }
                        else if(lastch == '=') {
                                getch();
                                lastst = asplus;
                                }
                        else lastst = plus;
                        break;
                case '-':
                        if(lastch == '-') {
                                getch();
                                lastst = autodec;
                                }
                        else if(lastch == '=') {
                                getch();
                                lastst = asminus;
                                }
                        else if(lastch == '>') {
                                getch();
                                lastst = pointsto;
                                }
                        else    lastst = minus;
                        break;
                case '*':
                        if(lastch == '=') {
                                getch();
                                lastst = astimes;
                                }
                        else lastst = star;
                        break;
                case '/':
                        if(lastch == '=') {
                                getch();
                                lastst = asdivide;
                                }
                        else if(lastch == '*') {
                                getch();
                                for(;;) {
                                        if(lastch == '*') {
                                                getch();
                                                if(lastch == '/') {
                                                        getch();
                                                        goto restart;
                                                        }
                                                }
                                        else
                                                getch();
                                        }
                                }
                        else lastst = divide;
                        break;
                case '^':
                        lastst = uparrow;
                        break;
                case ';':
                        lastst = semicolon;
                        break;
                case ':':
                        lastst = colon;
                        break;
                case '=':
                        if(lastch == '=') {
                                getch();
                                lastst = eq;
                                }
                        else lastst = assign;
                        break;
                case '>':
                        if(lastch == '=') {
                                getch();
                                lastst = geq;
                                }
                        else if(lastch == '>') {
                                getch();
                                if(lastch == '=') {
                                        getch();
                                        lastst = asrshift;
                                        }
                                else lastst = rshift;
                                }
                        else lastst = gt;
                        break;
                case '<':
                        if(lastch == '=') {
                                getch();
                                lastst = leq;
                                }
                        else if(lastch == '<') {
                                getch();
                                if(lastch == '=') {
                                        getch();
                                        lastst = aslshift;
                                        }
                                else lastst = lshift;
                                }
                        else lastst = lt;
                        break;
                case '\'':
                        ival = getsch();        /* get a string char */
                        if(lastch != '\'')
                                error(ERR_SYNTAX);
                        else
                                getch();
                        lastst = iconst;
                        break;
                case '\"':
                        for(i = 0;i < MAX_STRLEN;++i) {
                                if(lastch == '\"')
                                        break;
                                if((j = getsch()) == -1)
                                        break;
                                else
                                        laststr[i] = j;
                                }
                        laststr[i] = 0;
                        lastst = sconst;
                        if(lastch != '\"')
                                error(ERR_SYNTAX);
                        else
                                getch();
                        break;
                case '!':
                        if(lastch == '=') {
                                getch();
                                lastst = neq;
                                }
                        else lastst = not;
                        break;
                case '%':
                        if(lastch == '=') {
                                getch();
                                lastst = asmodop;
                                }
                        else lastst = modop;
                        break;
                case '~':
                        lastst = compl;
                        break;
                case '.':
                        lastst = dot;
                        break;
                case ',':
                        lastst = comma;
                        break;
                case '&':
                        if( lastch == '&') {
                                lastst = land;
                                getch();
                                }
                        else if( lastch == '=') {
                                lastst = asand;
                                getch();
                                }
                        else
                                lastst = and;
                        break;
                case '|':
                        if(lastch == '|') {
                                lastst = lor;
                                getch();
                                }
                        else if( lastch == '=') {
                                lastst = asor;
                                getch();
                                }
                        else
                                lastst = or;
                        break;
                case '(':
                        lastst = openpa;
                        break;
                case ')':
                        lastst = closepa;
                        break;
                case '[':
                        lastst = openbr;
                        break;
                case ']':
                        lastst = closebr;
                        break;
                case '{':
                        lastst = begin;
                        break;
                case '}':
                        lastst = end;
                        break;
                case '?':
                        lastst = hook;
                        break;
                default:
                        error(ERR_ILLCHAR);
                        goto restart;   /* get a real token */
                }
	}
        if(lastst == id)
                searchkw(0);
}

void needpunc(p)
enum e_sym      p;
{       if( lastst == p)
                getsym();
        else
                error(ERR_PUNCT);
}
