#include        <stdio.h>
#include        "c.h"
#include        "expr.h"
#include        "gen.h"
#include        "cglbdec.h"
#include        "proto.h"

/*
 *      TMS9900 C Compiler
 *
 *  Copyright 1990 by LGMA Products
 *
 *  This program is free for you to use for non-commercial purposes.
 *  It is, however, copyrighted so that commercial use is strictly
 *  forbidden.
 *
 *  Please contact:
 *
 *     LGMA PRODUCTS
 *     5618 AppleButter Hill Road
 *     Coopersburg, PA  18036
 *
 *  For suggestions, improvements, and the like.
 *
 *
/* Change History:
**
** Add new revisions to top.
**
**     Who           Date            Reason
**     ===           =========       =======================================
**     ALB           12/FEB/92       Fix problem with make_legal for direct
**                                   addressing modes (LI vs. MOV).
*/

/*
 *      This module contains misc. code generation "make" routines,
 *      really just an overflow from "gencode".
 */

struct amode    *makeassy(ptr) 
struct ascode   *ptr;
/*	make a reference to an assembly block of code */
{	struct  amode   *ap;
	ap = (struct amode *)xalloc(sizeof(struct amode));
	ap->mode = 0;
	ap->preg = 0;
	ap->offset = (struct enode *)ptr;
	return ap;
}

struct amode    *makedreg(r)
/*
 *      make an address reference to a data register.
 */
int     r;
{       struct amode    *ap;
        ap = (struct amode *)xalloc(sizeof(struct amode));
        ap->mode = am_reg;
        ap->preg = r;
        return ap;
}

struct amode    *make_immed(i)
/*
 *      make a node to reference an immediate value i.
 */
int     i;
{       struct amode    *ap;
        struct enode    *ep;
        ep = (struct enode *)xalloc(sizeof(struct enode));
        ep->nodetype = en_icon;
        ep->v.i = i;
        ap = (struct amode *)xalloc(sizeof(struct amode));
        ap->mode = am_immed;
        ap->offset = ep;
        return ap;
}

void   gen_loadr(ap, size, reg )
/*
 *  generate an immediate load based on the addressing mode in
 *  ap.  This gets to be a bit complicated on the TMS9900 as
 *  there are different forms for immediate loads, byte loads,
 *  etc.  This routine resolves these differences.
 */
struct amode    *ap;
int    size, reg;
{
    int value;
    struct enode *offset;

    offset = ap->offset;
    switch ( ap->mode ) {
      case am_immed:
        switch ( offset->nodetype ) {
        case en_icon:
          value = offset->v.i;
          if ( size == 1 ) {
            value = value << 8;
            gen_2code ( op_li, makedreg(reg), make_immed(value));
            }
          else
            gen_2code ( op_li, makedreg(reg), ap);
          break;
      
        default:
          gen_2code ( op_li,makedreg(reg), ap );
	  break;
        }
        break;

      default:
        gen_code ( op_mov, size, ap, makedreg(reg));
        break;
      }
}

void gen_load(ap, size)
/*
 *  generate an immediate load based on the addressing mode in
 *  ap.  This gets to be a bit complicated on the TMS9900 as
 *  there are different forms for immediate loads, byte loads,
 *  etc.  This routine resolves these differences.
 */
struct amode    *ap;
int    size;
{    gen_loadr ( ap, size, 8 );		/* use register 8 */
}

struct amode    *make_label(lab)
/*
 *      construct a reference node for an internal label number.
 */
int     lab;
{       struct enode    *lnode;
        struct amode    *ap;
        lnode = (struct enode *)xalloc(sizeof(struct enode));
        lnode->nodetype = en_labcon;
        lnode->v.i = lab;
        ap = (struct amode *)xalloc(sizeof(struct amode));
        ap->mode = am_direct;
        ap->offset = lnode;
        return ap;
}

struct amode    *make_labei(lab)
/*
 *      construct a reference node for an internal label number.
 */
int     lab;
{       struct enode    *lnode;
        struct amode    *ap;
        lnode = (struct enode *)xalloc(sizeof(struct enode));
        lnode->nodetype = en_labcon;
        lnode->v.i = lab;
        ap = (struct amode *)xalloc(sizeof(struct amode));
        ap->mode = am_immed;
        ap->offset = lnode;
        return ap;
}

struct amode    *make_offset(node)
/*
 *      make a direct reference to a node.
 */
struct enode    *node;
{       struct amode    *ap;
        ap = (struct amode *)xalloc(sizeof(struct amode));
        ap->mode = am_direct;
        ap->offset = node;
        return ap;
}
struct amode    *make_limme(node)
/*
 *      make an immediate reference to a node with a label
 */
struct enode    *node;
{       struct amode    *ap;
        ap = (struct amode *)xalloc(sizeof(struct amode));
        ap->mode = am_immed;
        ap->offset = node;
        return ap;
}
struct amode    *makeireg(r)
/*
 *      make an indirect address reference to a data register.
 */
int     r;
{       struct amode    *ap;
        ap = (struct amode *)xalloc(sizeof(struct amode));
        ap->mode = am_ind;
        ap->preg = r;
        return ap;
}

struct amode    *makeareg(r)
/*
 *      make an indirect address reference with autoincrement to the register
 */
int     r;
{       struct amode    *ap;
        ap = (struct amode *)xalloc(sizeof(struct amode));
        ap->mode = am_ainc;
        ap->preg = r;
        return ap;
}

struct amode    *make_direct(i)
/*
 *      make a direct reference to an immediate value.
 */
int     i;
{       return make_offset(makenode(en_icon,i,(struct enode *)0));
}

struct amode    *make_strlab(s)
/*
 *      generate a direct reference to a string label.
 */
char    *s;
{       struct amode    *ap;
        ap = (struct amode *)xalloc(sizeof(struct amode));
        ap->mode = am_direct;
        ap->offset =(struct enode *) makenode(en_nacon,(struct enode *)s,
                                                       (struct enode *)0);
        return ap;
}

int     natural_size(node)
/*
 *      return the natural evaluation size of a node.
 */
struct enode    *node;
{  int     siz0, siz1;
   if( node == 0 )
                return 0;
   switch( node->nodetype ) {
      case en_icon:
                          if( -128 <= node->v.i && node->v.i <= 127 )
                                return 1;
/*                        if( -32768 <= node->v.i && node->v.i <= 32767 )
                                return 2;
                        return 4;		*/
                return 2;
		case en_l_ref:
      case en_ul_ref:
		          return 4;
      case en_vfcall:
                          return 0;
      case en_fcall:
      case en_labcon:
      case en_nacon:
      case en_autocon:
      case en_tempref:
      case en_cbl:
      case en_cwl:
      case en_cbw:
      case en_w_ref:
      case en_uw_ref:
                return 2;
      case en_b_ref:
      case en_ub_ref:
                return 1;
		case en_v_ref:
			       return 0;
      case en_not:    case en_compl:
      case en_uminus: case en_assign:
      case en_ainc:   case en_adec:
                return natural_size(node->v.p[0]);
      case en_add:    case en_sub:
      case en_mul:    case en_div:
      case en_mod:    case en_and:
      case en_or:     case en_xor:
      case en_lsh:    case en_rsh:
      case en_rshl:
      case en_eq:     case en_ne:
      case en_lt:     case en_le:
      case en_gt:     case en_ge:
      case en_ult:    case en_ule:
      case en_ugt:    case en_uge:
      case en_land:   case en_lor:
      case en_asadd:  case en_assub:
      case en_asmul:  case en_asdiv:
      case en_asmod:  case en_asand:
      case en_asor:   case en_aslsh:
      case en_asrsh:
               siz0 = natural_size(node->v.p[0]);
               siz1 = natural_size(node->v.p[1]);
               if( siz1 > siz0 )
                     return siz1;
               else
                     return siz0;
      case en_void: 
      case en_cond:
               return natural_size(node->v.p[1]);
      default:
               printf("DIAG - natural size error.\n");
               break;
      }
      return 0;
}

void    make_legal(ap,flags,size)
/*
 *      make_legal will coerce the addressing mode in ap into a
 *      mode that is satisfactory for the flag word.
 */
struct amode    *ap;
int             flags, size;
{       struct amode    *ap2;

        if( ((flags & F_VOL) == 0) || ap->tempflag ) {
                switch( ap->mode ) {
                        case am_immed:
                                if( flags & F_IMMED )
                                        return;         /* mode ok */
                                break;
                        case am_reg:
                                if( flags & F_REG ) {
                                        return;
				}
                                break;
			case am_ainc:
                        case am_ind:
                        case am_direct:
			case am_indm:
                                if( flags & F_MEM )
                                        return;
                                break;
                        }
                }
        if( flags & F_REG ) {
                freeop(ap);             	/* maybe we can use it... */
                ap2 = (struct amode *) temp_data();	/* allocate to reg */
		if ( size == 4 ) {
		  (struct amode *)temp_data();		/* get two registers */
		}
                if ( ap->mode == am_immed ) {
		  gen_load ( ap, size );
		  gen_2code ( op_mov, makedreg(8), ap2 );
		  }
		else {
                  gen_code(op_mov,size,ap,ap2);
		}
                ap->mode = am_reg;
                ap->preg = ap2->preg;
                ap->deep = ap2->deep;
                ap->tempflag = 1;
                return;
                }

        if( size == 1 ) {
                freeop(ap);
                ap2 = (struct amode *)temp_data();
                gen_code(op_mov,1,ap,ap2);
		gen_2code(op_sla,ap2,make_immed(8));
                freeop(ap);
                ap->mode = ap2->mode;
                ap->preg = ap2->preg;
                ap->deep = ap2->deep;
                size = 2;
                }
        freeop(ap);
        ap2 = (struct amode *) temp_data();
        if ( ap->mode == am_immed ) {
          gen_code ( op_li, size, ap2, ap );
	}else{
          gen_code(op_mov,size,ap,ap2);
	}
        ap->mode = am_reg;
        ap->preg = ap2->preg;
        ap->deep = ap2->deep;
        ap->tempflag = 1;
}

void do_extend(ap,isize,osize,flags)
/*
 *      if isize is not equal to osize then the operand ap will be
 *      loaded into a register (if not already) and if osize is
 *      greater than isize it will be extended to match.
 */
struct amode    *ap;
int             isize, osize, flags;
{      	
	if( isize == osize )
                return;

        if( ap->mode != am_reg && ap->mode != am_reg )
                make_legal(ap,flags & F_REG,isize);

        switch( isize )
                {
                case 1:
		  gen_2code(op_sra,ap,make_immed(8));
		  break;

		case 2:
		  gen_2code(op_sla,ap,make_immed(8));
		  break;
                }
}

void swap_nodes(node)
/*
 *      exchange the two operands in a node.
 */
struct enode    *node;
{       struct enode    *temp;
        temp = node->v.p[0];
        node->v.p[0] = node->v.p[1];
        node->v.p[1] = temp;
}

void push_param(ep)
/*
 *      push the operand expression onto the stack.
 */
struct enode    *ep;
{       struct amode    *ap;

        ap = gen_expr(ep,F_ALL,2);

        if ( ap->mode == am_immed ) {
            gen_load  ( ap, 2 );
            gen_2code ( op_mov, makedreg(8), makeareg(14) );
        }else{
            gen_2code ( op_mov, ap, makeareg(14) );
	}

        freeop(ap);
}

struct amode    *gen_binary(node,flags,size,op)
/*
 *      generate code to evaluate a binary node and return 
 *      the addressing mode of the result.
 */
struct enode    *node;
int             flags, size, op;
{       struct amode    *ap1, *ap2;
        ap1 = gen_expr(node->v.p[0],F_VOL | F_REG,size);
        ap2 = gen_expr(node->v.p[1],F_ALL,size);
        validate(ap1);          /* in case push occurred */
	if ( ap2->mode == am_immed )
	  switch ( op ) {
	    case op_a:
	      if ( size == 1 ) {
		gen_load( ap2, size );
	        gen_code( op_a, size, makedreg(8), ap1 );
	      }else{
	        gen_code ( op_ai, size, ap1, ap2 );
	      }
	      break;

	    case op_s:
	      gen_load ( ap2, size );
	      gen_code ( op_s,  size, makedreg(8), ap1 );
	      break;

	}else{
       	    gen_code(op,size,ap2,ap1);
	}
        freeop(ap2);
        make_legal(ap1,flags,size);
        return ap1;
}

struct amode    *gen_index(node)
/*
 *      generate code to evaluate an index node (^+) and return
 *      the addressing mode of the result. This routine takes no
 *      flags since it always returns either am_ind or am_indx.
 */
struct enode    *node;
{       struct amode    *ap1, *ap2;

        ap1 = gen_expr(node->v.p[0], F_REG | F_IMMED,2);
        if( ap1->mode == am_immed ) {
                ap2 = gen_expr(node->v.p[1],F_REG,2);
                ap2->mode = am_indm;
                ap2->offset = ap1->offset;
                return ap2;
                }

        ap2 = gen_expr(node->v.p[1],F_ALL,2);   /* get right op */
        if( ap2->mode == am_immed  &&
	    ap1->mode == am_reg ) {		/* make am_indm */
                ap2->mode = am_indm;
                ap2->preg = ap1->preg;
                ap2->deep = ap1->deep;
                return ap2;
                }

        validate(ap1);
	if ( !(ap1->mode == am_reg && temp_register(ap1->preg) ) ) 
          make_legal(ap1,F_REG | F_VOL,2);
        gen_2code(op_a,ap2,ap1);            	/* add left to address reg */
        ap1->mode = am_ind;             	/* make indirect */
        freeop(ap2);                    	/* release any temps in ap2 */
        return ap1;                     	/* return indirect */
}

struct amode    *gen_shift(node,flags,size,op)
/*
 *      generate code to evaluate a shift node and return the
 *      address mode of the result.
 */
struct enode    *node;
int             flags, size, op;
{       struct amode    *ap1, *ap2;
        ap1 = gen_expr(node->v.p[0],F_REG | F_VOL,size);
/*        ap2 = gen_expr(node->v.p[1],F_REG | F_IMMED,1);     */
        ap2 = gen_expr(node->v.p[1], F_REG | F_IMMED, size );
        validate(ap1);
        if ( ap2->mode == am_reg ) {
          gen_2code ( op_mov, ap2, makedreg(0) );
          gen_2code ( op, ap1, makedreg(0) );
	}else{
          gen_2code(op,ap1,ap2);		/* size is always 2 */
	}
        freeop(ap2);
        make_legal(ap1,flags,size);
        return ap1;
}

