#include        <stdio.h>
#include        <string.h>
#include        <stdlib.h>
#include        "c.h"
#include        "expr.h"
#include        "gen.h"
#include        "cglbdec.h"
#include        "proto.h"

/*
 *      TMS9900 C compiler
 *
 *  Copyright 1990 by LGMA Products
 *
 *  This program is free for you to use for non-commercial purposes.
 *  It is, however, copyrighted so that commercial use is strictly
 *  forbidden.
 *
 *  Please contact:
 *
 *     LGMA PRODUCTS
 *     5618 AppleButter Hill Road
 *     Coopersburg, PA  18036
 *
 *  For suggestions, improvements, and the like.
 *
 */
#define CHUNKSIZE  8192		/* size of memory chunk to allocate */

/* the following routine swaps the memory usage between the peep list and
   the local/global memory lists.   In Amiga and IBM versions, this is
   essentially useless; for the GENEVE; this routine is replaced by an
   assembly language subroutine that actually swaps in and out 40k of
   memory that is used for the three lists; thereby extending the usable
   memory to approx 120k                                               */

#ifndef GENEVE
void swap(listnos)
int  listnos;
{  manlis = listnos;
}
#endif

/* xalloc - called to allocate some memory.  xalloc checks the "in-memory"
            list, and if memory is available; returns a pointer to that
            memory.   If memory isn't available, then calloc is called to
            fetch additional memory.   Each memory block contains a pointer
            to the next block for easy deallocation                       */

char    *xalloc(siz)
int     siz;
{       struct blkmem   *bp;
        char            *rv;
	if( siz & 1 )		/* if odd size */
		siz += 1;	/* make it even */

/* there are three types of memory allocation:
       a. global/local memory
       b. peepgen list memory
       c. assembly statement memory
                                                                        */
/* memory type 1 - PEEPLIST memory */

       if( manlis == PEEPLIST ) {
                if( mansize >= siz ) {
                        rv = &(manblk->m[manindx]);
                        mansize -= siz;
                        manindx += siz;
                        return rv;
                }else{
                        bp =(struct blkmem *)(calloc((unsigned)1,
                            (unsigned)CHUNKSIZE ));
			if( bp == NULL )
			{
				printf(" not enough extended memory.\n");
				closefiles();
				exit(1);
			}
                        bp->next = manblk;
                        manblk = bp;
                        mansize = CHUNKSIZE - (sizeof (struct blkmem)) - siz;
                        manindx = siz;
                        return manblk->m;
                        }
       }

/* memory type 2 - Assembly code list */

       if( manlis == ASMLIST ) {
                if( asmsize >= siz ) {
                        rv = &(asmblk->m[asmindx]);
                        asmsize -= siz;
                        asmindx += siz;
                        return rv;
                }else{
                        bp =(struct blkmem *)(calloc((unsigned)1,
                            (unsigned)CHUNKSIZE ));
			if( bp == NULL )
			{
				printf(" not enough extended memory.\n");
				closefiles();
				exit(1);
			}
                        bp->next = asmblk;
                        asmblk = bp;
                        asmsize = CHUNKSIZE - (sizeof (struct blkmem)) - siz;
                        asmindx = siz;
                        return asmblk->m;
                        }
	}

/* memory type 3 - Global / Local Memory */          
        if( global_flag ) {
                if( glbsize >= siz ) {
                        rv = &(glbblk->m[glbindx]);
                        glbsize -= siz;
                        glbindx += siz;
                        return rv;
                        }
                else    {
                        bp =(struct blkmem *)(calloc((unsigned)1,
                            (unsigned)CHUNKSIZE ));
			if( bp == NULL )
			{
				printf(" not enough memory.\n");
				closefiles();
				exit(1);
			}
                        bp->next = glbblk;
                        glbblk = bp;
                        glbsize = CHUNKSIZE - (sizeof (struct blkmem)) - siz;
                        glbindx = siz;
                        return glbblk->m;
                        }
                }
        else    {
                if( locsize >= siz ) {
                        rv = &(locblk->m[locindx]);
                        locsize -= siz;
                        locindx += siz;
                        return rv;
                        }
                else    {
                        bp =(struct blkmem *)(calloc((unsigned)1,
                            (unsigned)CHUNKSIZE ));
			if( bp == NULL )
			{
				printf(" not enough local memory.\n");
				closefiles();
				exit(1);
			}
                        bp->next = locblk;
                        locblk = bp;
                        locsize = CHUNKSIZE - (sizeof(struct blkmem)) - siz;
                        locindx = siz;
                        return locblk->m;
                        }
                }
}

/* release local memory, releases memory associated with lists of local
   variables, local definitions, code lists, peep lists, etc.            */
void release_local()
{       struct blkmem      *bp1, *bp2;
	swap (LOCALIST);
        bp1 = locblk;
        while( bp1 != 0 ) {
                bp2 = bp1->next;
                free( (char *) bp1 );
                bp1 = bp2;
                }
        locblk = 0;
        locsize = 0;
        lsyms.head = 0;

/* release all memory associated with PEEPLIST */
	swap ( PEEPLIST );
        bp1 = manblk;
        while( bp1 != 0 ) {
                bp2 = bp1->next;
                free( ( char *) bp1);
                bp1 = bp2;
                }
        manblk = 0;
        mansize = 0;

/* release all memory associated with the assembly source ASMLIST */
        swap ( ASMLIST );
        bp1  = asmblk;
        while ( bp1 != 0 ) {
		bp2 = bp1->next;
		free ( ( char *) bp1);
		bp1 = bp2;
	        }
	asmblk = 0;
	asmsize = 0;
	swap ( LOCALIST );
}

/* release global memory (end of compilation, releases memory associated
   with static variables, global definitions, etc.               */
void release_global()
{       struct blkmem      *bp1, *bp2;
        bp1 = glbblk;
        while( bp1 != 0 ) {
                bp2 = bp1->next;
                free( ( char *) bp1);
                bp1 = bp2;
                }
        glbblk = 0;
        glbsize = 0;
        gsyms.head = 0;         /* clear global symbol table */
        strtab = 0;             /* clear literal table */
}
