
#include        <stdio.h>
#include        "c.h"
#include        "expr.h"
#include        "gen.h"
#include        "cglbdec.h"
#include        "proto.h"

/*
 *      TMS9900 C compiler
 *
 *  Copyright 1990 by LGMA Products
 *
 *  This program is free for you to use for non-commercial purposes.
 *  It is, however, copyrighted so that commercial use is strictly
 *  forbidden.
 *
 *  Please contact:
 *
 *     LGMA PRODUCTS
 *     5618 AppleButter Hill Road
 *     Coopersburg, PA  18036
 *
 *  For suggestions, improvements, and the like.
 * 
 */

struct ocode    *iip;

struct amode    *copy_addr(ap)

/*
 *      copy an address mode structure (these things dont last).
 */
struct amode    *ap;
{       struct amode    locap;
	struct amode    *newap;

        if( ap == 0 )
                return 0;
/* copy operation is done in two steps; so that multiple memory spaces
   can be used to hold peeplist */
	swap ( LOCALIST );
        locap.mode     = ap->mode;
        locap.preg     = ap->preg;
        locap.tempflag = ap->tempflag;
        locap.deep     = ap->deep;
        locap.offset   = ap->offset;
 
/* now store structure in PEEPlist */
        swap ( PEEPLIST );
        newap = (struct amode *) xalloc(sizeof(struct amode));
        newap->mode     = locap.mode;
        newap->preg     = locap.preg;
        newap->tempflag = locap.tempflag;
        newap->deep     = locap.deep;
        newap->offset   = locap.offset;
        return newap;

}

void add_peep(new)
/*
 *      add the code reduction pointed to by new to the peep list.
 */
struct ocode    *new;
{       swap ( PEEPLIST );
        if( peep_head == 0 ) {
                peep_head = peep_tail = new;
                new->fwd = 0;
                new->back = 0;
        }else{
                new->fwd = 0;
                new->back = peep_tail;
                peep_tail->fwd = new;
                peep_tail = new;
                }
}

void gen_code(op,len,ap1,ap2)
/*
 *      generate a code sequence into the peep list.
 */
enum e_op op;
int             len;
struct amode    *ap1, *ap2;
{       struct ocode    *new;
        struct ocode    locnew;

        swap ( LOCALIST );
        locnew.opcode = op;
        locnew.length = len;
        locnew.oper1  = copy_addr(ap1);
        swap ( LOCALIST );
        locnew.oper2  = copy_addr(ap2);
        
        swap ( PEEPLIST );
        new = (struct ocode *) xalloc(sizeof(struct ocode));
        new->opcode = locnew.opcode;
        new->length = locnew.length;
        new->oper1  = locnew.oper1;
        new->oper2  = locnew.oper2;
        add_peep(new);
        swap ( LOCALIST );
}

void gen_1code ( op, ap1 )
/*
 *      generate a two byte code sequence in the peep list 
 */
enum e_op op;
struct amode  *ap1;
{	gen_code (op, 2, ap1, 0 );
}

void gen_2code ( op, ap1, ap2 )
/*
 *      generate a two byte code sequence in the peep list
 */
enum e_op op;
struct amode   *ap1, *ap2;
{	gen_code ( op, 2, ap1, ap2 );
}

void gen_label(labno)
/*
 *      add a compiler generated label to the peep list.
 */
int     labno;
{       struct ocode    *new;
        swap(PEEPLIST);
        new = (struct ocode *) xalloc(sizeof(struct ocode));
        new->opcode = op_label;
        new->oper1  = (struct amode *) labno;
        add_peep(new);
        swap(LOCALIST);
}

void put_ocode(p)
/*
 *      output the instruction passed.
 */
struct ocode    *p;
{       
	put_code(p->opcode,p->length,p->oper1,p->oper2);
}

int     equal_address(ap1,ap2)
/*c
 *      compare two address nodes and return true if they are
 *      equivalent.
 */
struct amode    *ap1, *ap2;
{       struct enode    *ep1, *ep2;
   
        if( ap1 == 0 || ap2 == 0 )
                return 0;
        if( ap1->mode != ap2->mode )
                return 0;
        switch( ap1->mode )
                {
                case am_reg:
                case am_ainc:
                case am_ind:
                        return ap1->preg == ap2->preg;

                case am_immed:
                        {
                        ep1 = ap1->offset;
                        ep2 = ap2->offset;
                        swap ( LOCALIST );
                        if ( ep1->nodetype != ep2->nodetype ) {
                           swap ( PEEPLIST );
                           return 0;
                           }
                        if ( ep1->nodetype != en_icon ) {
                           swap ( PEEPLIST );
                           return 0;
                           }
                        if ( ep1->v.i == ep2->v.i ) {
                           swap ( PEEPLIST );
                           return 1;
                        }else{
                           swap ( PEEPLIST );
                           return 0;
                           }
                        }
                case am_direct:
                        return ap1->offset == ap2->offset;

                case am_indm:
                       {
                        if ( ap1->preg != ap2->preg ) {
                           return 0;
                           }
                        ep1 = ap1->offset;
                        ep2 = ap2->offset;
                        swap ( LOCALIST );
                        if ( ep1->nodetype == ep2->nodetype  &&
                             ep1->nodetype == en_icon        &&
                             ep1->v.i      == ep2->v.i ) {
                           swap ( PEEPLIST );
                           return 1;
                        }else{
                           swap ( PEEPLIST );
                           return 0;
                           }
                        }
                 }
                 swap ( PEEPLIST );
                 return 0;
}
/*========================== used_reg =====================================*/
int used_reg ( reg, ip )
/*c used_reg -  is passed a register # and a pointer to the instruction 
                list.  used_reg returns TRUE if the passed register is
                utilized in the following sequence as a source register
                before:

                 1. conditional branch
                 2. unconditional branch
                 3. branch and link
                 4. destination register
                                                                          */
char reg;
struct ocode          *ip;
{
     while ( ip != 0 ) {

      if ( cond_op(ip) )
        return FALSE;

      if ( ip->opcode == op_bl ||
           ip->opcode == op_b  ||
           ip->opcode == op_jmp )
        return TRUE;

      if ( ip->opcode      == op_li &&
           ip->oper1->preg == reg )
        return TRUE;

      if ( ip->oper1->preg == reg )
        return FALSE;

      if ( ip->oper2         != 0    &&
           ip->oper2->preg   == reg )
        if ( ip->oper2->mode == am_reg &&
             ip->opcode      == op_mov ) 
           return TRUE;
        else
           return FALSE;

      ip = ip->fwd;
    }
    return FALSE;
}

/* ============================ removex ================================= */
/* removex - remove a node from the node list, by adjusting pointers.     */
/* ====================================================================== */
void removex ( ip )
struct  ocode  *ip;
{
struct  ocode  *prev;

	prev = ip->back;
	if ( prev != 0 ) {
          prev->fwd = ip->fwd;
          if( prev->fwd != 0 )
              prev->fwd->back = prev;
	  }

	optimized = TRUE;
}


/*========================== cond_op ======================================*/
int cond_op ( ip )
/*c  returns TRUE if current opcode is a conditional branch (e.g. jeq)      */
struct ocode           *ip;
{
        switch ( ip->opcode ) {
           case op_jeq:
           case op_jlt:
           case op_jgt:
           case op_jh:
           case op_jl:
           case op_jhe:
           case op_jle:
           case op_jne:
             return TRUE;
        }
        return FALSE;
}

/*=========================== peep_mov ===================================*/
void  peep_mov()
/*      peephole optimization for mov instructions.
 */
{	struct ocode    *prev, *nx, *nx1;

	prev = iip->back;
	nx   = iip->fwd;
	nx1  = nx->fwd;
	if ( nx1 == 0 || nx == 0 || prev == 0 ) {
	  return;
	}

/*====================================================================*/
/* combine instructions of type:
/*
/*          mov r4/5/6,r8,r1/2/3         a/s r4/5/6,x
/*          a/s r1/2/3,x                                              */

         if ( iip->oper1->mode == am_reg    &&
              iip->oper2->mode == am_reg    &&
              iip->oper2->preg == 8         &&
             (nx->opcode      == op_a      ||
              nx->opcode      == op_s  )   &&
             equal_address ( iip->oper2, nx->oper1 ) ) {
            nx->oper1->preg  = iip->oper1->preg;
            removex ( iip );
            return;
	 }
/*====================================================================*/
/* combine instruction of move of rx to a temp reg, followed by move
   of temp reg somewhere else                                         */

          if ( iip->oper2->mode == am_reg    &&
               temp_register ( iip->oper2->preg ) &&
               nx->opcode      == op_mov    &&
               equal_address ( iip->oper2, nx->oper1 ) ) {

            if ( cond_op(nx1) || nx1->opcode == op_bl ) {
               nx->oper1 = iip->oper1;
               removex ( iip );
               return;
	     }
	  }

/*====================================================================*/
/* delete moves to r8 if preceeded by another move, and followed by
   a conditional branch, i.e.:

         movb  *r1,*r2
         movb  *r1,r8
         jeq   x
									*/
         if ( nx->opcode == op_mov                   &&
              equal_address ( iip->oper1, nx->oper1 ) &&
              nx->oper2->mode == am_reg              &&
              nx->oper2->preg == 8                   &&
              cond_op ( nx1 ) ) {

           removex ( nx );
           return;
	   }

/*====================================================================*/
/*-  optimize:
             mov  r8,reg                mov  something,reg
             mov  something,reg                                       */

	    if ( iip->oper2->mode ==  am_reg  &&
		 nx->opcode      ==  op_mov  &&
		 equal_address ( iip->oper2, nx->oper2 ) &&
               ( nx->oper1->preg !=  iip->oper2->preg ) ) {

   	         removex ( iip );
	         return;
		 }

/*====================================================================*/
/* combine moves of r8 to temp register, i.e.:

             mov  r8,temp           mov  r8,somewhere
             mov  temp,somewhere                                     */


           if ( iip->oper1->mode == am_reg   &&     /* if 1st op == r8 */
                iip->oper1->preg == 8        &&     /* & second  == r1/2/3 */
                iip->oper2->mode == am_reg   &&
                temp_register ( iip->oper2->preg ) ) {

             if ( nx->opcode      == op_mov   &&
                  equal_address(iip->oper2, nx->oper1) &&
                  used_reg(iip->oper2->preg, nx1) ) {
                      iip->oper2 = nx->oper2;
                      removex(nx);
                      return;
		  }
		}

/*=====================================================================*/
/* make more efficient compares to 0, e.g.:

         movb *r4,r1             movb  *r4,r1
         clr  r8
         cb   r1,r8							*/

	if ( nx->opcode      ==  op_clr    &&
             nx->oper1->mode ==  am_reg    &&
             nx->oper1->preg ==  8         &&
             nx1->opcode     ==  op_c      &&
             equal_address ( nx->oper1, nx1->oper2 ) &&
             equal_address ( iip->oper2, nx1->oper1 ) )
          {
          removex ( nx );
          removex ( nx1 );
          return;
	  }

/*====================================================================*/
/* delete move of r8 to a register, followed by a statement that
   moves indirect the register to something else                      */

        if ( iip->oper1->mode == am_reg &&
             iip->oper1->preg == 8      &&
             iip->oper2->mode == am_reg &&
             temp_register ( iip->oper2->preg ) &&
             nx->opcode      == op_mov &&
             nx->oper1->mode == am_ind &&
             nx->oper1->preg == iip->oper2->preg ) {

          nx->oper1->preg = 8;
	  removex ( iip );
          iip = prev;
	  return;
	  }

/*====================================================================*/
/* delete moves of something to somewhere, and that somewhere back to
   the same something                                                */

         if ( nx->opcode == op_mov &&
	      equal_address ( iip->oper1, nx->oper2 ) &&
              equal_address ( iip->oper2, nx->oper1 ) ) {

	   removex ( nx );
	   return;
	   }

/*======================================================================*/
/* delete moves to myself, e.g. MOV 1,1                                 */
         if ( equal_address( iip->oper1, iip->oper2 ) &&
              !cond_op(nx)  ) {
           removex ( iip );
           return;
	 }

/*======================================================================*/
/* optimize:       mov tempreg,somewhere                                */
/*                 mov tempreg,tempreg                                  */
/*======================================================================*/
         if ( iip->oper1->mode == am_reg    &&
              temp_register(iip->oper1->preg)    &&
              nx->opcode      == op_mov    &&
              equal_address ( iip->oper1, nx->oper1 ) &&
              equal_address ( nx->oper1, nx->oper2 ) ) {
            removex ( nx );
            return;
	 }
}
/* ============================== replaceopcode ========================= */
/* replaceopcode - replaces the current node with the given opcode        */
/* ====================================================================== */
void replaceopcode ( opcode )
int    opcode;
{  
       iip->opcode = opcode;
       iip->oper2  = 0;
       optimized  = TRUE;
}

/*================================= replaceoperand ======================== */
/* replaceoperand - replaces the current node with the given opcode, and    */
/* uses the next instructions operand.  used in peep_li                     */
/* ======================================================================== */
void replaceoperand ( opcode )
int    opcode;
{   struct ocode *nx;
    nx = iip->fwd;
    iip->opcode = opcode;
    iip->oper1  = nx->oper2;
    iip->oper2  = 0;
    removex ( nx );
}

void  peep_li()
/*c
 *      peephole optimization for li instructions.
 *      changes various forms of load immediates to shorter sequences.
 *      this includes substitution of clr & seto, as well as use
 *      of inc, inct, dec, dect for incrementing/decrementing.
 */
{       struct enode    *ep;
	struct ocode    *nx, *nx1;
	int             peepval;

        ep = iip->oper2->offset;
	swap ( LOCALIST );
        if( ep->nodetype != en_icon ) {
          return;
        }
	peepval = ep->v.i;

	swap ( PEEPLIST );
	nx = iip->fwd;		/* get next node address */

/*  optimize forms of li r8,const, mov r8,@x    */
        if ( nx->opcode == op_mov     &&
	     equal_address(nx->oper1,iip->oper1) &&
	     nx->length == 2  ) {

/*  optimize  li r8,0; mov r8,@x          */
	    if ( peepval == 0 ) {
              replaceoperand ( op_clr );
	      return;
	    }

/* optimize  li r8,-1; mov r8,@x          */
	    if ( peepval == -1 ) {
              replaceoperand ( op_seto );
	      return;
	    }

/* optimize li r8,x; mov r8,@y, li r8,x; mov r8,@y, etc. */
	    nx = nx->fwd;
            for ( ;; ) {
	      if ( nx->opcode == op_li    &&
	          equal_address(nx->oper1,iip->oper1) &&
	          equal_address(nx->oper2,iip->oper2) ) {
	          nx1 = nx;
	          nx  = nx->fwd;
	          removex(nx1);
	          if ( nx->opcode == op_mov &&
                    equal_address(nx->oper1,iip->oper1) &&
                    nx->length == 2 ) {
                    nx = nx->fwd;
	          }else{
                    return;
	          }
               }else{
                  break;
	    }
	 }

/* optimize li r8,x; mov r8,tempreg                   */
          nx = iip->fwd;
          if ( nx->oper2->mode == am_reg    &&
	       iip->oper1->preg == 8 ) {
             iip->oper1->preg = nx->oper2->preg;
             removex ( nx );
             return;
	  }
       }

/* optimize forms of li r8,y;  a r8,@x    */
	nx = iip->fwd;
	if ( (nx->opcode == op_a) &&
	     (equal_address(nx->oper1,iip->oper1) ) &&
             (nx->length == 2 ) ) {

          switch ( peepval ) {
             case 1:
               replaceoperand ( op_inc );
               break;
             case 2:
               replaceoperand ( op_inct );
               break;
             case -1:
               replaceoperand ( op_dec );
               break;
             case -2:
               replaceoperand ( op_dect );
               break;
	  }
          return;
	}

/* optimize forms of  li r8,const;  s r8,@x    */
	if ( nx->opcode == op_s &&
	     equal_address(nx->oper1,iip->oper1)  &&
             nx->length == 2  ) {

          switch ( peepval ) {
             case 1:
               replaceoperand ( op_dec );
               break;
             case 2:
               replaceoperand ( op_dect );
               break;
             case -1:
               replaceoperand ( op_inc );
               break;
             case -2:
               replaceoperand ( op_inct );
               break;
	  }
          return;
        }

        switch ( peepval ) {
          case 0:
            replaceopcode ( op_clr );
            return;
          case -1:
            replaceopcode ( op_seto );
	    return;
	}
}
void    peep_ci()
/*
 *      peephole optimization for compare instructions.
 *      changes compare to 0 to mov reg,reg, and if previous
 *      instruction should have set the condition codes properly
 *      delete.  return value is true if instruction was deleted.
 */
{       struct ocode    *prev;
        struct enode    *ep;
	int		peepval;

        ep = iip->oper2->offset;
	swap ( LOCALIST );
	peepval = ep->v.i;
        if( ep->nodetype == en_icon ) {
	  swap ( PEEPLIST );
          if ( peepval == 0 ) {
            optimized = TRUE;
            iip->oper2 = iip->oper1;
            iip->opcode = op_mov;
            prev = iip->back;
            if( prev == 0 )
                return;

/* if condition checks match, then delete compare instruction */

            if( (( (prev->opcode == op_mov) &&
                equal_address(prev->oper1,iip->oper1)) &&
                prev->oper2->mode != am_reg) ||
                (prev->opcode != op_label &&
                equal_address(prev->oper2,iip->oper1)) ) 
                {
                prev->fwd = iip->fwd;
                if( prev->fwd != 0 )
                        prev->fwd->back = prev;
                }
	  }
	}
}

void   peep_ai()
/*
 *   peep_ai  -  add immediate optimization, use inc/dec instructions
 *               for obvious ai.
 */
{
        struct enode    *ep;
	int		peepval;

        ep = iip->oper2->offset;
	swap ( LOCALIST );
        if( ep->nodetype != en_icon ) {
		swap ( PEEPLIST );
		return;
	}
	peepval = ep->v.i;
	swap ( PEEPLIST );
        switch ( peepval ) {
          case 1:
            replaceopcode ( op_inc );
	    return;
          case 2:
            replaceopcode ( op_inct );
	    return;
          case -1:
            replaceopcode ( op_dec );
            return;
          case -2:
            replaceopcode ( op_dect );
            return;
        }
        return;
}

int   det_length ( ip )
/*c     det_length - given the node ip, determines the number of bytes
 *                  the instruction will generate.  Used to determine
 *                  if short or long branches can be used in the
 *                  expression.
 */
struct  ocode      *ip;
{
int  length;

 	if ( ip->opcode == op_label )
          return 0;

	if ( ip->opcode == op_data )
          return (int) ip->length;

	length = 2;			/* set basic length */

        if ( cond_op ( ip ) || ip->opcode == op_jmp )
          return length;

	switch ( ip->oper1->mode ) {
	  case am_immed:
          case am_direct:
          case am_indm:
	    length += 2;
	    break;
	  }

        if ( ip->oper2 != 0 ) {
	  switch ( ip->oper2->mode ) {
	  case am_immed:
          case am_direct:
          case am_indm:
	    length +=2;
	    break;
	  }
	}
	return length;
}

int   search_lab ( ip )
/*c
 *     search_lab - searches the node list for the specified label,
 *                  both backwards and forwards.  if the node is
 *                  beyond addressing range (I have to guess in
 *                  forward direction) then return 0; otherwise
 *                  return range.
 */
struct  ocode         *ip;
{	
int	bytefwd, bytebak, labno;
struct  ocode         *nx;
struct  enode         *ep1;
int		      peepval;

	bytefwd = 2;
	bytebak = 2;
	nx      = ip->fwd;
	ep1     = ip->oper1->offset;		/* get label number */
	swap ( LOCALIST );
	peepval = ep1->v.i;
	swap ( PEEPLIST );

	while ( bytefwd < 128 && nx != 0 ) {
	  bytefwd += det_length( nx );

	  if ( nx->opcode == op_label ) {
	    labno = (int) nx->oper1;
            if(   labno == peepval )
              return bytefwd;
	    }
	  nx = nx->fwd;
	  }

	nx = ip->back;
	while ( bytebak < 128 && nx != 0 ) {
	  bytebak += det_length ( nx );

	  if ( nx->opcode == op_label ) {
	    labno = (int) nx->oper1;
            if(   labno == peepval )
              return bytebak;
	    }
	  nx = nx->back;
	  }

	return 0;		/* label not found */
}

void  peep_incdec()
/*
 *   optimizations for inc, dec, inct, dect
 */
{
struct ocode    *nx, *nx1;

	nx  = iip->fwd;
	nx1 = nx->fwd;
	if ( nx1 == 0 )
	  return;

        if ( nx->opcode    ==  op_mov               &&
	     equal_address ( iip->oper1, nx->oper1 ) &&
	     nx->oper2->mode == am_reg              &&
	     nx->oper2->preg == 8                   &&
             cond_op ( nx1 ) ) {

	  removex ( nx );
	  return;
	  }
}

void  peep_uctran()
/*c
 *      peephole optimization for unconditional transfers.
 */
{       struct ocode *nx;
	int          peepval;
	struct enode *ep1;

/* delete unconditional transfers if no label follows (useless) */
	ep1 = iip->oper1->offset;
	while( iip->fwd != 0 && iip->fwd->opcode != op_label ) {
                iip->fwd = iip->fwd->fwd;
                if( iip->fwd != 0 )
                        iip->fwd->back = iip;
                optimized = TRUE;
                return;
                }
/*  change b @x to jmp x if within range		*/

	if ( iip->opcode      == op_b               &&
             iip->oper1->mode == am_direct ) {
             swap ( LOCALIST );
	     if ( ep1->nodetype == en_labcon ) {
	       swap ( PEEPLIST );
               if ( search_lab ( iip ) ) {
                 iip->opcode = op_jmp;
	         iip->oper1->mode = am_immed;
                 optimized = TRUE;
	       }
	     }
	  }
/* delete branches if just branching to label following (useless) */

	swap ( PEEPLIST );
        nx = iip->fwd;
	if ( nx         != 0          &&
             nx->opcode == op_label )     {

            swap ( LOCALIST );
	    peepval = ep1->v.i;
	    swap ( PEEPLIST );
	    if ( peepval == (int)nx->oper1 ) {
              removex( iip );
              return;
	    }
	}
}

void  peep_ctran ( )
/*
 *   peep_ctran - optimizer for conditional branch instructions.
 *                looks for sequences of conditional branches followed
 *                by long branch that can be optimized into single branch.
 */
{
struct    ocode     *nx;

	nx   =   iip->fwd;
	if ( nx->opcode == op_b && search_lab(nx) ) {
	  switch ( iip->opcode ) {
	    case op_jeq:
		removex ( iip );
		nx->opcode = op_jne;
		nx->oper1->mode = am_immed;
		break;
	    case op_jne:
		removex ( iip );
		nx->opcode = op_jeq;
		nx->oper1->mode = am_immed;
		break;
	    case op_jgt:
		iip->opcode = op_jlt;
		iip->oper1  = nx->oper1;
		nx->oper1->mode = am_immed;
		nx->opcode = op_jeq;
		break;
	    case op_jlt:
		iip->opcode = op_jgt;
		iip->oper1  = nx->oper1;
		nx->oper1->mode = am_immed;
		nx->opcode = op_jeq;
		break;
	    case op_jh:
		removex(iip);
		nx->opcode = op_jle;
		nx->oper1->mode = am_immed;
		break;
	    case op_jhe:
		removex(iip);
		nx->opcode = op_jl;
		nx->oper1->mode = am_immed;
		break;
	    case op_jl:
		removex(iip);
		nx->opcode = op_jhe;
		nx->oper1->mode = am_immed;
		break;
	    case op_jle:
		removex(iip);
		nx->opcode = op_jh;
		nx->oper1->mode = am_immed;
		break;
	    }
	  }
}

/* =========================== extract_label ========================== */
/* if operation is label constant, then return label #; else -1         */
/* ==================================================================== */
int extract_label ( ap )
struct amode *ap;
{   struct enode    *ep;
    int             reflabel;

    if ( ap == 0 ) {
      return -1;
    }
    
    if ( ap->mode == am_immed || ap->mode == am_direct ) {
      ep = ap->offset;
      swap ( LOCALIST );
      if ( ep->nodetype == en_labcon ) {
        reflabel = ep->v.i;
        swap ( PEEPLIST );
        return reflabel;
      }
    }
    swap ( PEEPLIST );
    return -1;
}

/*
 *	Label optimization:
 *         remove labels which aren't referenced
 */
void opt_labels()
{
	struct ocode	*ip;
	struct ocode	*ip1;
	int target_label;

	for (ip=peep_head; ip != 0; ip = ip->fwd) {

		if (ip->opcode != op_label)   continue;

/* remove unreferenced labels */
		target_label = (int)ip->oper1;

		for (ip1 = peep_head; ip1 != 0; ip1 = ip1->fwd) {
                     if ( extract_label(ip1->oper1) == target_label ) {
                        goto next_instruct;
                     }

                     if ( extract_label(ip1->oper2) == target_label ) {
                        goto next_instruct;
                     }
		}
                removex ( ip );

next_instruct: ;
	}
}

void opt3()
/*
 *      peephole optimizer. This routine calls the instruction
 *      specific optimization routines above for each instruction
 *      in the peep list.
 *
 *      Optimizations are repeated on the module node list until no
 *      more optimizations can be found.   Typically this is two passes.
 */
{       
	int             passes;

	passes    = 0;
        optimized = TRUE;
        while ( optimized && passes < 4 ) {
          passes++;
          optimized = FALSE;
          swap ( PEEPLIST );
	  iip = peep_head;
          while( iip != 0 ) {
            switch( iip->opcode ) {
                    case op_li:
                            peep_li();
                            break;
                    case op_ci:
                            peep_ci();
                            break;
                    case op_ai:
                            peep_ai();
                            break;
                    case op_mov:
                            peep_mov();
                            break;
                    case op_b:
                    case op_jmp:
                            peep_uctran();
                            break;
                    case op_inc:
                    case op_inct:
                    case op_dec:
                    case op_dect:
                            peep_incdec();
                            break;
                    case op_jne:
                    case op_jeq:
                    case op_jlt:
                    case op_jgt:
                    case op_jh:
                    case op_jl:
                    case op_jhe:
                    case op_jle:
                            peep_ctran();
                            break;
                    }
            swap ( PEEPLIST );
            iip = iip->fwd;
            }
/*            if ( optimized ) opt_labels(); */
	}
}

void opt5()
/*
 *      opt5 - stack coalescing.  search for a bl, when found,
 *             try to perform stack coalescing (combining this
 *             stack instruction with ones following)
 *
 *             Note that the stack pops need to be combined very
 *             carefully.  As soon as a branch or label or BL that
 *             is not a subr jump is reached; no furthur combination is
 *             possible.  Also, the subr jumps must be at the same level
 *             to be combined; e.g. the following code sequence could
 *             get into trouble:
 *
 *                       y = val ( jval(x), z );
 *
 *             as the calls to jval & val are not at the same level.  This
 *             routine monitors all stack pushes and pops; and will only
 *             do combinations if at level 0.
                                                                       */

{       struct ocode    *ip, *nx, *nx1, *last;
	struct enode    *ep1, *ep2;
	int level;

	swap ( PEEPLIST );
	ip = peep_head;
	level = 0;
        last  = 0;

/*	do for each instruction in peep list */

	for ( ;ip != 0; ip = ip->fwd ) {

/*  switch on opcode type */
	    switch ( ip->opcode )  {
	      case op_mov:			/* stack push */
		if ( ip->oper2->mode == am_ainc &&
                     ip->oper2->preg == 14 ) {
                   level += 2;
		     }
                break;

              case op_dect:			/* stack pop */
                if ( ip->oper1->preg == 14 ) {
                  level -= 2;
		}
                break;
						/* branch and link */
              case op_bl:
	        nx = ip->fwd;
	        if ( nx == 0 )
                   break;
	        nx1 = nx->fwd;
	        if ( nx1 == 0 )
                   break;
						/* see if bl call subr */
                if (  ip->oper1->mode == am_ind  &&
                      ip->oper1->preg == 12      &&
		      nx->opcode      == op_data ) {

                    if ( nx1->opcode     == op_ai   &&
                         nx1->oper1->preg== 14         ) {

                         ip = nx1;
			 ep1    = ip->oper2->offset;
			 swap ( LOCALIST );
			 level += ep1->v.i;
			 swap ( PEEPLIST );
                         if ( level == 0 ) {
						/* if level is 0; then
						    sequence can begin    */
                            if ( last != 0 ) {
				ep1 = ip->oper2->offset;
				ep2 = last->oper2->offset;
				swap ( LOCALIST );
				ep1->v.i += ep2->v.i;
				swap ( PEEPLIST );
                                removex(last);
		            }
                            last = ip;
			 }          		/* if level == 0 */
			 }          		/* if nx1->opcode == op_ai */
		      }else{       		/* if nx->opcode == op_data */
                         level = 0;
                         last  = 0;
                      }
	              break;

                case op_b:			/* any type of branch */
                case op_jmp:
                case op_label:
		case op_jeq:
		case op_jlt:
		case op_jgt:
		case op_jh:
		case op_jl:
		case op_jhe:
		case op_jle:
		case op_jne:
                      level = 0;
                      last  = 0;
                      break;

		default:
                      break;
	    }						/* end switch */
	}						/* end while  */
}

void flush_peep()
/*
 *      output all code and labels in the peep list.
 */
{       swap ( PEEPLIST );
        if ( optflag ) {
            opt5();        	/* do stack coalescing */
            opt3();           	/* do the peephole optimizations */
        }
        while( peep_head != 0 ) {
                if( peep_head->opcode == op_label ) {
                        put_label(peep_head->oper1);
                }else{
                        put_ocode(peep_head);
		}
                peep_head = peep_head->fwd;
        }
        swap ( LOCALIST );
}
