#include        <stdio.h>
#include        <string.h>
#include        "c.h"
#include        "expr.h"
#include        "gen.h"
#include        "cglbdec.h"
#include        "proto.h"

/*
 *      TMS9900 C Compiler
 *
 *  Copyright 1990 by LGMA Products
 *
 *  This program is free for you to use for non-commercial purposes.
 *  It is, however, copyrighted so that commercial use is strictly
 *  forbidden.
 *
 *  Please contact:
 *
 *     LGMA PRODUCTS
 *     5618 AppleButter Hill Road
 *     Coopersburg, PA  18036
 *
 *  For suggestions, improvements, and the like.
 *
 */

int     isalnum(c)
char    c;
{       return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') ||
                (c >= '0' && c <= '9');
}

int     isidch(c)
char    c;
{       return isalnum(c) || c == '_' || c == '$';
}

int     isspace(c)
char    c;
{       return c == ' ' || c == '\t' || c == '\n';
}

int     isdigit(c)
char    c;
{       return (c >= '0' && c <= '9');
}

/* ============================= shrink ============================== */
/* shrink - shrink a pre-processor token by removing leading blanks,   */
/*          embedded comments, and trailing blanks.                    */
/*          This is done to reduce storage requirements during comp-   */
/*          ilation, a problem on Geneve and MSDOS versions.           */
/* =================================================================== */
void     shrink ( to, from )
char   *to, *from;
{    char   *next;

     while ( *from == ' ' ) {		/* skip any leading blanks */
       from++;
     }
     while ( *from != 0 ) {
       next = from;
       next++;
/* if comment line found, then strip it out */
       if ( *from == '/' && *next == '*' ) {
         from++;
         from++;
         while ( *from != 0 ) {
           next = from;
           next++;
           if ( *from == '*' && *next == '/' ) {
             from++;
             from++;
             break;
           }
           from++;
	 }
       }else{        
         *to++ = *from++;
       }
     }
/* strip any trailing blanks */
     to--;
     while ( *to == '\n' || *to == ' ' ) {
       *to--;
     }
     to++;
     *to++ = '\n';
     *to = 0;       
}

/* ============================= preprocess ========================== */
/* preprocess - handle a preprocessor (#) command                      */
/* =================================================================== */
int  preprocess()
{       ++lptr;
        lastch = ' ';
        getsym();               /* get first word on line */
        searchkw(1);
/* if currently not processing an IFDEF */
        if ( ifdepth < 0 ) {
          switch ( lastst ) {
            case kw_include:
                return doinclude();
            case kw_define:
                return dodefine();
            case kw_ifdef:
                return doifdef();
            case kw_ifndef:
                return doifndef();
	    case kw_asm:
		return doasm();
	    case kw_endasm:
		return doendasm();
            default:
                error(ERR_PREPROC);
                return 0;
	  }
/* if currently processing an IFDEF & processing mode is TRUE */
	}else if ( ifarray[ifdepth] == TRUE ) {
          switch ( lastst ) {
            case kw_include:
                return doinclude();
            case kw_define:
                return dodefine();
            case kw_ifdef:
                return doifdef();
            case kw_ifndef:
                return doifndef();
            case kw_else:
                return doelse();
            case kw_endif:
                return doendif();
	    case kw_asm:
		return doasm();
	    case kw_endasm:
		return doendasm();
            default:
                error ( ERR_PREPROC );
                return;
	  }
/* if currently processing an IFDEF and processing mode is FALSE */
	}else{
          switch ( lastst ) {
             case kw_else:
                return doelse();
             case kw_endif:
                return doendif();
             default:
                return 0;
	  }
	}
}

/* ============================= doinclude =========================== */
/*                                                                     */
/*   process a #include statement.  file names in form of ".." and     */
/*   <..> are supported.                                               */
/*                                                                     */
/*                                                                     */
int  doinclude()
{       int     typest;
	char    *filxptr;

	typest = incname();		/* get file to include */
        if( lastst != sconst ) {
                error(ERR_INCLFILE);
                return;
                }
        inclline[incldepth]   = lineno;
        inclfile[incldepth++] = input;  /* push current input file */
	input = 0;			/* put old file on hold */
	if ( !typest )			/* quoted string */
	  input = fopen ( laststr, "r" );

	if ( input == 0 ) {		/* if <> form of include name */
	  incptr = inc_files;		/* start at include pointer */

	  while ( (*incptr != 0) && ( input == 0 ) ) {
	    filxptr  = incfile;
	    for ( ;*filxptr++ = *incptr++; )
	      {;}
	    filxptr--;
	    strcpy ( filxptr, laststr );
	    filxptr  = incfile;
	    input = fopen( filxptr, "r" );
            if ( input != 0 )
              printf("Processing %s\n",filxptr);
	    }
          }else{
              printf("Processing %s\n",laststr);
	  }

        if( input == 0 ) {
                input = inclfile[--incldepth];
                error(ERR_CANTOPEN);
                }
        return TRUE;
}
/* =========================== dodefine ================================ */
/* dodefine - process the #define statement                              */
/* ===================================================================== */
int  dodefine()
{       SYM     *sp;
        char    su[100];
        
        getsym();               /* get past #define */
        if( lastst != id ) {
                error(ERR_DEFINE);
                return;
                }
        ++global_flag;          /* always do #define as globals */
        sp =(SYM *) xalloc(sizeof(SYM));
        sp->name =(char *)litlate(lastid);
        shrink( su, lptr );
        sp->value.s = (char *) litlate(&su);
        insert(sp,&defsyms);
        --global_flag;
        return TRUE;
}

/* =========================== doifdef  ================================ */
/* doifdef - process the #ifdef statement                                */
/* ===================================================================== */
int  doifdef()
{	SYM     *sp;
	getsym();		/* get past #ifdef */
        sp =(SYM *)search(lastid,&defsyms);

	if ( ++ifdepth > 10 ) {
          	--ifdepth;
	  	error(ERR_DEPTH);
		return;
	}

	if ( sp == 0 ) {
	  ifarray[ifdepth] = FALSE;
	}else{
	  ifarray[ifdepth] = TRUE;
	}
	getch();
        return TRUE;
}

/* =========================== doifndef ================================ */
/* doifndef - process the #ifdef statement                               */
/* ===================================================================== */
int  doifndef()
{	SYM     *sp;
	getsym();		/* get past #ifdef */
        sp =(SYM *)search(lastid,&defsyms);

	if ( ++ifdepth > 10 ) {
          	--ifdepth;
	  	error(ERR_DEPTH);
		return;
	}

	if ( sp == 0 ) {
	  ifarray[ifdepth] = TRUE;
	}else{
	  ifarray[ifdepth] = FALSE;
	}
	getch();
        return TRUE;
}

/* =========================== doendif  ================================ */
/* doendif - process the #endif statement                                */
/* ===================================================================== */
int   doendif()
{
	if ( ifdepth < 0 ) {
		error(ERR_ENDIF);
		return;
	}
        --ifdepth;
	return TRUE;
}

/* =========================== doelse   ================================ */
/* doelse - process the #else statement                                  */
/* ===================================================================== */
int   doelse()
{
	if ( ifdepth < 0 ) {
		error(ERR_ENDIF);
		return;
	}
	ifarray[ifdepth] = !ifarray[ifdepth];
	return TRUE;
}

/* =========================== doasm =================================== */
/* doasm - include assembly language statements                          */
/* ===================================================================== */
int   doasm()
{	
	if ( assembly ) {
	  error ( ERR_ENDIF );
	  return TRUE;
	}else{
	  assembly = TRUE;
	  return FALSE;
	}
}

/* =========================== doendasm ================================ */
/* doendasm - include assembly language statements                       */
/* ===================================================================== */
int  doendasm ()
{    	if ( assembly ) {
	  return FALSE;
	}else{
	  error ( ERR_ENDIF );
	  return TRUE;
	}
}

/* =========================== incname  ================================ */
/* incname - parse a string for "include" statement                      */
/* ===================================================================== */
/*
 *  incname is called to parse the include file name.  This file name
 *  can have one of two constructs:
 *
 *       " file_name "
 *            or
 *       < file_name >
 *  
 */
int     incname()
{       register int    i, j, typest;

        while(isspace(lastch))
                getch();

        if( lastch == -1)
                lastst = kw_eof;

	if ( lastch == '\"' )
	  typest  = FALSE;		/* indicates normal string */

	else if ( lastch == '<' )
	  typest  = TRUE;		/* indicates <> string */

	else
	  error ( ERR_SYNTAX );

	getch();
        for(i = 0;i < MAX_STRLEN; ++i) {
           if(lastch == '\"' && !typest )
               break;
	   if(lastch == '>'  &&  typest )
	       break;
           if((j = getsch()) == -1)
               break;
           else
               laststr[i] = j;
           }
        laststr[i] = 0;
        lastst = sconst;
	if ( lastch == '\"' && !typest )
	  getch();
	else if ( lastch == '>'  &&  typest )
	  getch();
	else
	  error(ERR_SYNTAX);
	return typest;
/*	return TRUE;		*/
}
