#include        <stdio.h>
#include        "c.h"
#include        "expr.h"
#include        "gen.h"
#include        "cglbdec.h"
#include        "proto.h"

/*
 *      TMS9900 C Compiler
 *
 *  Copyright 1990 by LGMA Products
 *
 *  This program is free for you to use for non-commercial purposes.
 *  It is, however, copyrighted so that commercial use is strictly
 *  forbidden.
 *
 *  Please contact:
 *
 *     LGMA PRODUCTS
 *     5618 AppleButter Hill Road
 *     Coopersburg, PA  18036
 *
 *  For suggestions, improvements, and the like.
 *
 */

/*
 *      this module handles the allocation and de-allocation of
 *      temporary registers. when a temporary register is allocated
 *      the stack depth is saved in the field deep of the address
 *      mode structure. when validate is called on an address mode
 *      structure the stack is popped until the register is restored
 *      to it's pre-push value.
 */

#define  NO_REGISTERS   3

int  reglist[]  = {1,   2, 3};
int  pairlist[] = {11, 12, 0};		/* paired list (for longs) */

/* temp_register - return TRUE if register is a temporary register */
int  temp_register ( registerno )
int  registerno;
{    	int   i;
	for ( i=0; i<=NO_REGISTERS; i++ ) {
	   if ( reglist[i] == registerno ) {
	      return 1;
	   }
	   if ( pairlist[i] == registerno ) {
	      return 1;
	   }
	}
	return 0;
}

void gen_push(reg,rmode)
/*
 *      this routine generates code to push a register onto the stack
 */
int     reg;
enum e_am rmode;
{       struct amode    *ap1;
        ap1 = (struct amode *) xalloc(sizeof(struct amode));
        ap1->preg = reg;
        ap1->mode = rmode;
        gen_2code(op_mov,ap1,makeareg(14));
	  
}

void gen_pop(reg,rmode)
/*
 *      generate code to pop the primary register in ap from the
 *      stack.
 */
int     reg;
enum e_am rmode;
{       struct amode    *ap1;
        ap1 = (struct amode *) xalloc(sizeof(struct amode));
        ap1->preg = reg;
        ap1->mode = rmode;
	gen_1code(op_dect,makedreg(14));
	gen_2code(op_mov, makeireg(14), ap1);
}

void initstack()
/*
 *      this routine should be called before each expression is
 *      evaluated to make sure the stack is balanced and all of
 *      the registers are marked free.
 */
{       next_data = 0;
        max_data  = NO_REGISTERS-1;
}

void  validate(ap)
/*
 *      validate will make sure that if a register within an address
 *      mode has been pushed onto the stack that it is popped back
 *      at this time.
 */
struct amode    *ap;
{       switch( ap->mode )
                {
                case am_direct:
                case am_immed:
                        return;         	/* no registers used */
                default:
			if ( !temp_register(ap->preg) ) {
                                return; 	/* not a temporary */
                              }
                        if( max_data - ap->deep >= NO_REGISTERS ) {
                                gen_pop(ap->preg,am_reg);
                                --max_data;
                                }
                        break;
                }
}

struct amode    *temp_data()
/*
 *      allocate a temporary data register and return it's
 *      addressing mode.
 */
{       int  registerno;

	struct amode    *ap;
        ap = (struct amode *) xalloc(sizeof(struct amode));
        ap->mode = am_reg;
	registerno = next_data % NO_REGISTERS;
        ap->preg = reglist[registerno];
        ap->deep = next_data;
        if( next_data > max_data )
                {
		registerno = next_data % NO_REGISTERS;
                gen_push( reglist[registerno], am_reg);
                max_data = next_data;
                }
        ++next_data;
        return ap;
}

void  freeop(ap)
/*
 *      release any temporary registers used in an addressing mode.
 */
struct amode    *ap;
{       if( ap->mode == am_immed || ap->mode == am_direct )
                return;         /* no registers used */
	if( temp_register ( ap->preg ) ) {
		--next_data;
	}
}
