#include        <stdio.h>
#include        <string.h>
#include        "c.h"
#include        "expr.h"
#include        "gen.h"
#include        "cglbdec.h"
#include        "proto.h"

/*
 *      TMS9900 C Compiler
 *
 *  Copyright 1990 by LGMA Products
 *
 *  This program is free for you to use for non-commercial purposes.
 *  It is, however, copyrighted so that commercial use is strictly
 *  forbidden.
 *
 *  Please contact:
 *
 *     LGMA PRODUCTS
 *     5618 AppleButter Hill Road
 *     Coopersburg, PA  18036
 *
 *  For suggestions, improvements, and the like.
 *
 */

/*
 *      the statement module handles all of the possible c statements
 *      and builds a parse tree of the statements.
 *
 *      each routine returns a pointer to a statement parse node which
 *      reflects the statement just parsed.
 */

struct snode    *asmstmt()
/*
 *      asmstmt parses the #asm statement
*/
{       struct snode     *snp;
	struct amode  *ap;

	snp = (struct snode *)xalloc(sizeof(struct snode));
	snp->stype = st_asm;
	
	snp->s1 = (struct snode *)asmread();
	getsym();
	return snp;
 }

struct ascode *asmread()
  {
	struct ascode *asbegin, *asptr, *asnew;
	char   *ptr;

	swap ( LOCALIST );
	asbegin = 0;
	asptr   = 0;
	for ( ;; ) {
	  if ( inpline[0] == '#' ) {     /* if #endasm being processed */
	    getch();                    /* skip past the prev stmt */
	    getch();                    /* skip past the # */
	    getid();                    /* get the next id */
	    searchkw(1);                /* search pre-process keywords */
	    if ( lastst == kw_endasm ) {
	      break;
	    }else{
	      error ( ERR_SYNTAX );     /* else, syntax error */
	      break;
	    }
	  }
	  swap ( ASMLIST );
	  asnew = (struct ascode *)xalloc ( sizeof ( struct ascode ) );
	  if ( asbegin == 0 ) {
	    asbegin = asnew;
	    asptr   = asnew;
	  }else{
	    asptr->fwd = asnew;
	  }
	  asnew->fwd = 0;
	  ptr = xalloc ( strlen ( inpline ) + 1 );
	  strcpy ( ptr, inpline );
	  asnew->stmt = ptr;
	  asptr       = asnew;
	  swap ( LOCALIST );
	  if ( get_line() ) {
	    break;
	  }
	}

/* when finished, must create a code node and link to the current code
   chain */
	assembly = FALSE;
	getsym();
	return asbegin;
}

struct snode    *whilestmt()
/*
 *      whilestmt parses the c while statement.
 */
{       struct snode    *snp;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	snp->stype = st_while;
	getsym();
	if( lastst != openpa )
		error(ERR_EXPREXPECT);
	else    {
		getsym();
		if( expression(&(snp->exp)) == 0 )
			error(ERR_EXPREXPECT);
		needpunc( closepa );
		snp->s1 = statement();
		}
	return snp;
}

struct snode    *dostmt()
/*
 *      dostmt parses the c do-while construct.
 */
{       struct snode    *snp;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	snp->stype = st_do;
	getsym();
	snp->s1 = statement();
	if( lastst != kw_while )
		error(ERR_WHILEXPECT);
	else    {
		getsym();
		if( lastst != openpa )
			error(ERR_EXPREXPECT);
		else    {
			getsym();
			if( expression(&(snp->exp)) == 0 )
				error(ERR_EXPREXPECT);
			needpunc(closepa);
			}
		if( (lastst != end) && (lastst != kw_eof) )
			needpunc( semicolon );
		}
	return snp;
}

struct snode    *forstmt()
{       struct snode    *snp;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	getsym();
	needpunc(openpa);
	if( expression( &(                 snp->label )) == 0 )
		snp->label = 0;
	needpunc(semicolon);
	snp->stype = st_for;
	if( expression(&(snp->exp)) == 0 )
		snp->exp = 0;
	needpunc(semicolon);
	if( expression(&(snp->s2)) == 0 )
		snp->s2 = 0;
	needpunc(closepa);
	snp->s1 = statement();
	return snp;
}

struct snode    *ifstmt()
/*
 *      ifstmt parses the c if statement and an else clause if
 *      one is present.
 */
{       struct snode    *snp;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	snp->stype = st_if;
	getsym();
	if( lastst != openpa )
		error(ERR_EXPREXPECT);
	else    {
		getsym();
		if( expression(&(snp->exp)) == 0 )
			error(ERR_EXPREXPECT);
		needpunc( closepa );
		snp->s1 = statement();
		if( lastst == kw_else ) {
			getsym();
			snp->s2 = statement();
			}
		else
			snp->s2 = 0;
		}
	return snp;
}

struct snode    *casestmt()
/*
 *      cases are returned as seperate statements. for normal
 *      cases label is the case value and s2 is zero. for the
 *      default case s2 is nonzero.
 */
{       struct snode    *snp;
	struct snode    *head, *tail;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	if( lastst == kw_case ) {
		getsym();
		snp->s2 = 0;
		snp->stype = st_case;
		snp->label =(int *) constexpr();
		}
	else if( lastst == kw_default) {
		getsym();
		snp->s2 =(struct snode *) 1;
		snp->stype = st_default;          /* type */
		}
	else    {
		error(ERR_NOCASE);
		return 0;
		}
	needpunc(colon);
	head = 0;
	while(  lastst != end &&
		lastst != kw_case &&
		lastst != kw_default &&
		lastst != kw_eof ) {
		if( head == 0 ) {
			head = tail = statement();
			while ( tail->next != 0 ) {
			  tail = tail->next;
			}
		}else{
			tail->next = statement();
			if( tail->next != 0 )
				tail = tail->next;
			}
		tail->next = 0;
		}
	snp->s1 = head;
	return snp;
}

int     checkcases(head)
/*
 *      checkcases will check to see if any duplicate cases
 *      exist in the case list pointed to by head.
 */
struct snode    *head;
{
	struct snode    *top, *cur;
	cur = top = head;
	while( top != 0 )
	{
		cur = top->next;
		while( cur != 0 )
		{
			if( (!(cur->s1 || cur->s2) && cur->label == top->label)
				|| (cur->s2 && top->s2) )
			{
		    printf(" DIAG - duplicate case label %d\n",(int)cur->label);
				return 1;
			}
			cur = cur->next;
		}
		top = top->next;
	}
	return 0;
}

struct snode    *switchstmt()
{       struct snode    *snp;
	struct snode    *head, *tail;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	snp->stype = st_switch;
	getsym();
	needpunc(openpa);
	if( expression(&(snp->exp)) == 0 )
		error(ERR_EXPREXPECT);
	needpunc(closepa);
	needpunc(begin);
	head = 0;
	while( lastst != end && lastst != kw_eof ) {
		if( head == 0 )  {
			head = tail = casestmt();
			while ( tail->next != 0 ) {
			  tail = tail->next;
			}
		}else{
			tail->next = casestmt();
			if( tail->next != 0 )
				tail = tail->next;
			}
		tail->next = 0;
		}
	snp->s1 = head;
	getsym();
	if( checkcases(head) )
		error(ERR_DUPCASE);
	return snp;
}

struct snode    *retstmt()
{       struct snode    *snp;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	snp->stype = st_return;
	getsym();
	expression(&(snp->exp));
	if( lastst != end && lastst != kw_eof )
		needpunc( semicolon );
	return snp;
}

struct snode    *breakstmt()
{       struct snode    *snp;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	snp->stype = st_break;
	getsym();
	if( lastst != end && lastst != kw_eof )
		needpunc( semicolon );
	return snp;
}

struct snode    *contstmt()
{       struct snode    *snp;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	snp->stype = st_continue;
	getsym();
	if( lastst != end && lastst != kw_eof )
		needpunc( semicolon );
	return snp;
}

struct snode    *exprstmt()
/*
 *      exprstmt is called whenever a statement does not begin
 *      with a keyword. the statement should be an expression.
 */
{       struct snode    *snp;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	snp->stype = st_expr;
	if( expression(&(snp->exp)) == 0 ) {
		error(ERR_EXPREXPECT);
		getsym();
		}
	if( lastst != end && lastst != kw_eof )
		needpunc( semicolon );
	return snp;
}

struct snode    *compound()
/*
 *      compound processes a block of statements and forms a linked
 *      list of the statements within the block.
 *
 *      compound expects the input pointer to already be past the
 *      begin symbol of the block.
 */
{       struct snode    *head, *tail;
	head = 0;
	while( lastst != end && lastst != kw_eof ) {
		if( head == 0 ) {
			head = tail = statement();
			while ( tail->next != 0 ) {
			  tail = tail->next;
			}
		}else{
			tail->next = statement();
			if( tail->next != 0 )
				tail = tail->next;
			}
		}
	getsym();
	return head;
}

struct snode    *labelstmt()
/*
 *      labelstmt processes a label that appears before a
 *      statement as a seperate statement.
 */
{       struct snode    *snp;
	SYM             *sp;
	snp = (struct snode *)xalloc(sizeof(struct snode));
	snp->stype = st_label;
	if( (sp =(SYM *) search(lastid,lsyms.head)) == 0 ) {
		sp =(SYM *) xalloc(sizeof(SYM));
		sp->name =(char *) litlate(lastid);
		sp->storage_class = sc_label;
		sp->tp = 0;
		sp->value.i = (int) nextlabel++;
		insert(sp,&lsyms);
		}
	else    {
		if( sp->storage_class != sc_ulabel )
			error(ERR_LABEL);
		else
			sp->storage_class = sc_label;
		}
	getsym();       /* get past id */
	needpunc(colon);
	if( sp->storage_class == sc_label ) {
		snp->label = (int *)(long)sp->value.i;
		snp->next = 0;
		return snp;
		}
	return 0;
}

struct snode    *gotostmt()
/*
 *      gotostmt processes the goto statement and puts undefined
 *      labels into the symbol table.
 */
{       struct snode    *snp;
	SYM             *sp;
	getsym();
	if( lastst != id ) {
		error(ERR_IDEXPECT);
		return 0;
		}
	snp =(struct snode *) xalloc(sizeof(struct snode));
	if( (sp =(SYM *) search(lastid,lsyms.head)) == 0 ) {
		sp =(SYM *) xalloc(sizeof(SYM));
		sp->name =(char *) litlate(lastid);
		sp->value.i = nextlabel++;
		sp->storage_class = sc_ulabel;
		sp->tp = 0;
		insert(sp,&lsyms);
		}
	getsym();       /* get past label name */
	if( lastst != end && lastst != kw_eof )
		needpunc( semicolon );
	if( sp->storage_class != sc_label && sp->storage_class != sc_ulabel)
		error( ERR_LABEL );
	else    {
		snp->stype = st_goto;
		snp->label =(int *)(long)sp->value.i;
		snp->next = 0;
		return snp;
		}
	return 0;
}

struct snode    *statement()
/*
 *      statement figures out which of the statement processors
 *      should be called and transfers control to the proper
 *      routine.
 */
{       struct snode    *snp;
	switch( lastst ) {
		case semicolon:
			getsym();
			snp = 0;
			break;
		case begin:
			getsym();
			snp = compound();
			return snp;
		case kw_asm:
			snp = asmstmt();
			break;
		case kw_if:
			snp = ifstmt();
			break;
		case kw_while:
			snp = whilestmt();
			break;
		case kw_for:
			snp = forstmt();
			break;
		case kw_return:
			snp = retstmt();
			break;
		case kw_break:
			snp = breakstmt();
			break;
		case kw_goto:
			snp = gotostmt();
			break;
		case kw_continue:
			snp = contstmt();
			break;
		case kw_do:
			snp = dostmt();
			break;
		case kw_switch:
			snp = switchstmt();
			break;
		case id:
			while( isspace(lastch) )
				getch();
			if( lastch == ':' )
				return labelstmt();
			/* else fall through to process expression */
		default:
			snp = exprstmt();
			break;
		}
	if( snp != 0 )
		snp->next = 0;
	return snp;
}
